##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Library for forking storage server and connecting client storage"""


import logging
import random
import socket
import time

import ZODB.tests.util
import zope.testing.setupstack

from ZEO import _forker
from ZEO._compat import WIN


logger = logging.getLogger('ZEO.tests.forker')

DEBUG = _forker.DEBUG

ZEOConfig = _forker.ZEOConfig


def encode_format(fmt):
    # The list of replacements mirrors
    # ZConfig.components.logger.handlers._control_char_rewrites
    for xform in (("\n", r"\n"), ("\t", r"\t"), ("\b", r"\b"),
                  ("\f", r"\f"), ("\r", r"\r")):
        fmt = fmt.replace(*xform)
    return fmt


runner = _forker.runner
stop_runner = _forker.stop_runner
start_zeo_server = _forker.start_zeo_server

if WIN:
    def _quote_arg(s):
        return '"%s"' % s
else:
    def _quote_arg(s):
        return s

shutdown_zeo_server = _forker.shutdown_zeo_server


def get_port(ignored=None):
    """Return a port that is not in use.

    Checks if a port is in use by trying to connect to it.  Assumes it
    is not in use if connect raises an exception. We actually look for
    2 consective free ports because most of the clients of this
    function will use the returned port and the next one.

    Raises RuntimeError after 10 tries.
    """

    for _i in range(10):
        port = random.randrange(20000, 30000)
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s1 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            try:
                s.connect(('127.0.0.1', port))
            except OSError:
                pass  # Perhaps we should check value of error too.
            else:
                continue

            try:
                s1.connect(('127.0.0.1', port+1))
            except OSError:
                pass  # Perhaps we should check value of error too.
            else:
                continue

            return port

        finally:
            s.close()
            s1.close()
    raise RuntimeError("Can't find port")


def can_connect(port):
    c = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        try:
            c.connect(('127.0.0.1', port))
        except OSError:
            return False  # Perhaps we should check value of error too.
        else:
            return True
    finally:
        c.close()


def setUp(test):
    ZODB.tests.util.setUp(test)

    servers = []

    def start_server(storage_conf=None, zeo_conf=None, port=None, keep=False,
                     addr=None, path='Data.fs', protocol=None, blob_dir=None,
                     suicide=True, debug=False, **kw):
        """Start a ZEO server.

        Return the server and admin addresses.
        """
        if port is None:
            if addr is None:
                port = 0
            else:
                port = addr[1]
        elif addr is not None:
            raise TypeError("Can't specify port and addr")
        addr, stop = start_zeo_server(
            storage_conf=storage_conf,
            zeo_conf=zeo_conf,
            port=port,
            keep=keep,
            path=path,
            protocol=protocol,
            blob_dir=blob_dir,
            suicide=suicide,
            debug=debug,
            **kw)
        servers.append(stop)
        return addr, stop

    test.globs['start_server'] = start_server

    def stop_server(stop):
        stop()
        servers.remove(stop)

    test.globs['stop_server'] = stop_server

    def cleanup_servers():
        for stop in list(servers):
            stop()

    zope.testing.setupstack.register(test, cleanup_servers)

    test.globs['wait_until'] = wait_until
    test.globs['wait_connected'] = wait_connected
    test.globs['wait_disconnected'] = wait_disconnected


def wait_until(label=None, func=None, timeout=3, onfail=None):
    if label is None:
        if func is not None:
            label = func.__name__
    elif not isinstance(label, str) and func is None:
        func = label
        label = func.__name__

    if func is None:
        def wait_decorator(f):
            wait_until(label, f, timeout, onfail)

        return wait_decorator

    giveup = time.time() + timeout
    while not func():
        if time.time() > giveup:
            if onfail is None:
                raise AssertionError("Timed out waiting for: ", label)
            else:
                return onfail()
        time.sleep(0.01)


def wait_connected(storage):
    wait_until("storage is connected", storage.is_connected)


def wait_disconnected(storage):
    wait_until("storage is disconnected",
               lambda: not storage.is_connected())


debug_logging = _forker.debug_logging
whine = _forker.whine
ThreadlessQueue = _forker.ThreadlessQueue



"""Test layer for threaded-server tests

uvloop currently has a bug,
https://github.com/MagicStack/uvloop/issues/39, that causes failure if
multiprocessing and threaded servers are mixed in the same
application, so we isolate the few threaded tests in their own layer.
"""
import ZODB.tests.util


threaded_server_tests = ZODB.tests.util.MininalTestLayer(
    'threaded_server_tests')



#! /usr/bin/env python
##############################################################################
#
# Copyright (c) 2001-2005 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""
Cache simulation.


Note:

- The simulation isn't perfect.

- The simulation will be far off if the trace file
  was created starting with a non-empty cache
"""

import argparse
import bisect
import random
import re
import struct
import sys
# we assign ctime locally to facilitate test replacement!
from time import ctime

from ZODB.utils import z64

import ZEO.cache

from ..cache import ZEC_HEADER_SIZE
from .cache_stats import add_interval_argument
from .cache_stats import add_tracefile_argument


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    # Parse options.
    MB = 1 << 20
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--size", "-s",
                        default=20*MB, dest="cachelimit",
                        type=lambda s: int(float(s)*MB),
                        help="cache size in MB (default 20MB)")
    add_interval_argument(parser)
    parser.add_argument("--rearrange", "-r",
                        default=0.8, type=float,
                        help="rearrange factor")
    add_tracefile_argument(parser)

    simclass = CircularCacheSimulation

    options = parser.parse_args(args)

    f = options.tracefile
    interval_step = options.interval

    # Create simulation object.
    sim = simclass(options.cachelimit, options.rearrange)
    interval_sim = simclass(options.cachelimit, options.rearrange)

    # Print output header.
    sim.printheader()

    # Read trace file, simulating cache behavior.
    f_read = f.read
    unpack = struct.unpack
    FMT = ">iiH8s8s"
    FMT_SIZE = struct.calcsize(FMT)
    assert FMT_SIZE == 26

    last_interval = None
    while 1:
        # Read a record and decode it.
        r = f_read(FMT_SIZE)
        if len(r) < FMT_SIZE:
            break
        ts, code, oidlen, start_tid, end_tid = unpack(FMT, r)
        if ts == 0:
            # Must be a misaligned record caused by a crash; skip 8 bytes
            # and try again.  Why 8?  Lost in the mist of history.
            f.seek(f.tell() - FMT_SIZE + 8)
            continue
        oid = f_read(oidlen)
        if len(oid) < oidlen:
            break
        # Decode the code.
        dlen, version, code = ((code & 0x7fffff00) >> 8,
                               code & 0x80,
                               code & 0x7e)
        # And pass it to the simulation.
        this_interval = int(ts) // interval_step
        if this_interval != last_interval:
            if last_interval is not None:
                interval_sim.report()
                interval_sim.restart()
                if not interval_sim.warm:
                    sim.restart()
            last_interval = this_interval
        sim.event(ts, dlen, version, code, oid, start_tid, end_tid)
        interval_sim.event(ts, dlen, version, code, oid, start_tid, end_tid)

    f.close()
    # Finish simulation.
    interval_sim.report()
    sim.finish()


class Simulation:
    """Base class for simulations.

    The driver program calls: event(), printheader(), finish().

    The standard event() method calls these additional methods:
    write(), load(), inval(), report(), restart(); the standard
    finish() method also calls report().
    """

    def __init__(self, cachelimit, rearrange):
        self.cachelimit = cachelimit
        self.rearrange = rearrange
        # Initialize global statistics.
        self.epoch = None
        self.warm = False
        self.total_loads = 0
        self.total_hits = 0       # subclass must increment
        self.total_invals = 0     # subclass must increment
        self.total_writes = 0
        if not hasattr(self, "extras"):
            self.extras = (self.extraname,)
        self.format = self.format + " %7s" * len(self.extras)
        # Reset per-run statistics and set up simulation data.
        self.restart()

    def restart(self):
        # Reset per-run statistics.
        self.loads = 0
        self.hits = 0       # subclass must increment
        self.invals = 0     # subclass must increment
        self.writes = 0
        self.ts0 = None

    def event(self, ts, dlen, _version, code, oid,
              start_tid, end_tid):
        # Record first and last timestamp seen.
        if self.ts0 is None:
            self.ts0 = ts
            if self.epoch is None:
                self.epoch = ts
        self.ts1 = ts

        # Simulate cache behavior.  Caution:  the codes in the trace file
        # record whether the actual cache missed or hit on each load, but
        # that bears no necessary relationship to whether the simulated cache
        # will hit or miss.  Relatedly, if the actual cache needed to store
        # an object, the simulated cache may not need to (it may already
        # have the data).
        action = code & 0x70
        if action & 0x20:
            # Load.
            self.loads += 1
            self.total_loads += 1
            # Asserting that dlen is 0 iff it's a load miss.
            # assert (dlen == 0) == (code in (0x20, 0x24))
            self.load(oid, dlen, start_tid, code)
        elif action & 0x40:
            # Store.
            assert dlen
            self.write(oid, dlen, start_tid, end_tid)
        elif action & 0x10:
            # Invalidate.
            self.inval(oid, start_tid)
        elif action == 0x00:
            # Restart.
            self.restart()
        else:
            raise ValueError("unknown trace code 0x%x" % code)

    def write(self, oid, size, start_tid, end_tid):
        pass

    def load(self, oid, size, start_tid, code):
        # Must increment .hits and .total_hits as appropriate.
        pass

    def inval(self, oid, start_tid):
        # Must increment .invals and .total_invals as appropriate.
        pass

    format = "%12s %6s %7s %7s %6s %6s %7s"

    # Subclass should override extraname to name known instance variables;
    # if extraname is 'foo', both self.foo and self.total_foo must exist:
    extraname = "*** please override ***"

    def printheader(self):
        print("{}, cache size {} bytes".format(self.__class__.__name__,
                                           addcommas(self.cachelimit)))
        self.extraheader()
        extranames = tuple([s.upper() for s in self.extras])
        args = ("START TIME", "DUR.", "LOADS", "HITS",
                "INVALS", "WRITES", "HITRATE") + extranames
        print(self.format % args)

    def extraheader(self):
        pass

    nreports = 0

    def report(self):
        if not hasattr(self, 'ts1'):
            return
        self.nreports += 1
        args = (ctime(self.ts0)[4:-8],
                duration(self.ts1 - self.ts0),
                self.loads, self.hits, self.invals, self.writes,
                hitrate(self.loads, self.hits))
        args += tuple([getattr(self, name) for name in self.extras])
        print(self.format % args)

    def finish(self):
        # Make sure that the last line of output ends with "OVERALL".  This
        # makes it much easier for another program parsing the output to
        # find summary statistics.
        print('-'*74)
        if self.nreports < 2:
            self.report()
        else:
            self.report()
            args = (
                ctime(self.epoch)[4:-8],
                duration(self.ts1 - self.epoch),
                self.total_loads,
                self.total_hits,
                self.total_invals,
                self.total_writes,
                hitrate(self.total_loads, self.total_hits))
            args += tuple([getattr(self, "total_" + name)
                           for name in self.extras])
            print(self.format % args)


# For use in CircularCacheSimulation.
class CircularCacheEntry:
    __slots__ = (
        # object key: an (oid, start_tid) pair, where start_tid is the
        # tid of the transaction that created this revision of oid
        'key',

        # tid of transaction that created the next revision; z64 iff
        # this is the current revision
        'end_tid',

        # Offset from start of file to the object's data record; this
        # includes all overhead bytes (status byte, size bytes, etc).
        'offset',
        )

    def __init__(self, key, end_tid, offset):
        self.key = key
        self.end_tid = end_tid
        self.offset = offset


class CircularCacheSimulation(Simulation):
    """Simulate the ZEO 3.0 cache."""

    # The cache is managed as a single file with a pointer that
    # goes around the file, circularly, forever.  New objects
    # are written at the current pointer, evicting whatever was
    # there previously.

    extras = "evicts", "inuse"

    evicts = 0

    def __init__(self, cachelimit, rearrange):
        Simulation.__init__(self, cachelimit, rearrange)
        self.total_evicts = 0  # number of cache evictions

        # Current offset in file.
        self.offset = ZEC_HEADER_SIZE

        # Map offset in file to (size, CircularCacheEntry) pair, or to
        # (size, None) if the offset starts a free block.
        self.filemap = {ZEC_HEADER_SIZE: (self.cachelimit - ZEC_HEADER_SIZE,
                                          None)}
        # Map key to CircularCacheEntry.  A key is an (oid, tid) pair.
        self.key2entry = {}

        # Map oid to tid of current revision.
        self.current = {}

        # Map oid to list of (start_tid, end_tid) pairs in sorted order.
        # Used to find matching key for load of non-current data.
        self.noncurrent = {}

        # The number of overhead bytes needed to store an object pickle
        # on disk (all bytes beyond those needed for the object pickle).
        self.overhead = ZEO.cache.allocated_record_overhead

        # save evictions so we can replay them, if necessary
        self.evicted = {}

    def restart(self):
        Simulation.restart(self)
        if self.evicts:
            self.warm = True
        self.evicts = 0
        self.evicted_hit = self.evicted_miss = 0

    evicted_hit = evicted_miss = 0

    def load(self, oid, size, tid, code):
        if (code == 0x20) or (code == 0x22):
            # Trying to load current revision.
            if oid in self.current:  # else it's a cache miss
                self.hits += 1
                self.total_hits += 1

                tid = self.current[oid]
                entry = self.key2entry[(oid, tid)]
                offset_offset = self.offset - entry.offset
                if offset_offset < 0:
                    offset_offset += self.cachelimit
                    assert offset_offset >= 0

                if offset_offset > self.rearrange * self.cachelimit:
                    # we haven't accessed it in a while.  Move it forward
                    size = self.filemap[entry.offset][0]
                    self._remove(*entry.key)
                    self.add(oid, size, tid)

            elif oid in self.evicted:
                size, e = self.evicted[oid]
                self.write(oid, size, e.key[1], z64, 1)
                self.evicted_hit += 1
            else:
                self.evicted_miss += 1

            return

        # May or may not be trying to load current revision.
        cur_tid = self.current.get(oid)
        if cur_tid == tid:
            self.hits += 1
            self.total_hits += 1
            return

        # It's a load for non-current data.  Do we know about this oid?
        L = self.noncurrent.get(oid)
        if L is None:
            return  # cache miss
        i = bisect.bisect_left(L, (tid, None))
        if i == 0:
            # This tid is smaller than any we know about -- miss.
            return
        lo, hi = L[i-1]
        assert lo < tid
        if tid > hi:
            # No data in the right tid range -- miss.
            return
        # Cache hit.
        self.hits += 1
        self.total_hits += 1

    # (oid, tid) is in the cache.  Remove it:  take it out of key2entry,
    # and in `filemap` mark the space it occupied as being free.  The
    # caller is responsible for removing it from `current` or `noncurrent`.
    def _remove(self, oid, tid):
        key = oid, tid
        e = self.key2entry.pop(key)
        pos = e.offset
        size, _e = self.filemap[pos]
        assert e is _e
        self.filemap[pos] = size, None

    def _remove_noncurrent_revisions(self, oid):
        noncurrent_list = self.noncurrent.get(oid)
        if noncurrent_list:
            self.invals += len(noncurrent_list)
            self.total_invals += len(noncurrent_list)
            for start_tid, end_tid in noncurrent_list:
                self._remove(oid, start_tid)
            del self.noncurrent[oid]

    def inval(self, oid, tid):
        if tid == z64:
            # This is part of startup cache verification:  forget everything
            # about this oid.
            self._remove_noncurrent_revisions(oid)

        if oid in self.evicted:
            del self.evicted[oid]

        cur_tid = self.current.get(oid)
        if cur_tid is None:
            # We don't have current data, so nothing more to do.
            return

        # We had current data for oid, but no longer.
        self.invals += 1
        self.total_invals += 1
        del self.current[oid]
        if tid == z64:
            # Startup cache verification:  forget this oid entirely.
            self._remove(oid, cur_tid)
            return

        # Our current data becomes non-current data.
        # Add the validity range to the list of non-current data for oid.
        assert cur_tid < tid
        L = self.noncurrent.setdefault(oid, [])
        bisect.insort_left(L, (cur_tid, tid))
        # Update the end of oid's validity range in its CircularCacheEntry.
        e = self.key2entry[oid, cur_tid]
        assert e.end_tid == z64
        e.end_tid = tid

    def write(self, oid, size, start_tid, end_tid, evhit=0):
        if end_tid == z64:
            # Storing current revision.
            if oid in self.current:  # we already have it in cache
                if evhit:
                    raise ValueError('WTF')
                return
            self.current[oid] = start_tid
            self.writes += 1
            self.total_writes += 1
            self.add(oid, size, start_tid)
            return
        if evhit:
            raise ValueError('WTF')
        # Storing non-current revision.
        L = self.noncurrent.setdefault(oid, [])
        p = start_tid, end_tid
        if p in L:
            return  # we already have it in cache
        bisect.insort_left(L, p)
        self.writes += 1
        self.total_writes += 1
        self.add(oid, size, start_tid, end_tid)

    # Add `oid` to the cache, evicting objects as needed to make room.
    # This updates `filemap` and `key2entry`; it's the caller's
    # responsibilty to update `current` or `noncurrent` appropriately.
    def add(self, oid, size, start_tid, end_tid=z64):
        key = oid, start_tid
        assert key not in self.key2entry
        size += self.overhead
        avail = self.makeroom(size+1)   # see cache.py
        e = CircularCacheEntry(key, end_tid, self.offset)
        self.filemap[self.offset] = size, e
        self.key2entry[key] = e
        self.offset += size
        # All the space made available must be accounted for in filemap.
        excess = avail - size
        if excess:
            self.filemap[self.offset] = excess, None

    # Evict enough objects to make at least `need` contiguous bytes, starting
    # at `self.offset`, available.  Evicted objects are removed from
    # `filemap`, `key2entry`, `current` and `noncurrent`.  The caller is
    # responsible for adding new entries to `filemap` to account for all
    # the freed bytes, and for advancing `self.offset`.  The number of bytes
    # freed is the return value, and will be >= need.
    def makeroom(self, need):
        if self.offset + need > self.cachelimit:
            self.offset = ZEC_HEADER_SIZE
        pos = self.offset
        while need > 0:
            assert pos < self.cachelimit
            size, e = self.filemap.pop(pos)
            if e:   # there is an object here (else it's already free space)
                self.evicts += 1
                self.total_evicts += 1
                assert pos == e.offset
                _e = self.key2entry.pop(e.key)
                assert e is _e
                oid, start_tid = e.key
                if e.end_tid == z64:
                    del self.current[oid]
                    self.evicted[oid] = size-self.overhead, e
                else:
                    L = self.noncurrent[oid]
                    L.remove((start_tid, e.end_tid))
            need -= size
            pos += size
        return pos - self.offset  # total number of bytes freed

    def report(self):
        self.check()
        free = used = total = 0
        for size, e in self.filemap.values():
            total += size
            if e:
                used += size
            else:
                free += size

        self.inuse = round(100.0 * used / total, 1)
        self.total_inuse = self.inuse
        Simulation.report(self)
        # print self.evicted_hit, self.evicted_miss

    def check(self):
        oidcount = 0
        pos = ZEC_HEADER_SIZE
        while pos < self.cachelimit:
            size, e = self.filemap[pos]
            if e:
                oidcount += 1
                assert self.key2entry[e.key].offset == pos
            pos += size
        assert oidcount == len(self.key2entry)
        assert pos == self.cachelimit

    def dump(self):
        print(len(self.filemap))
        L = list(self.filemap)
        L.sort()
        for k in L:
            v = self.filemap[k]
            print(k, v[0], repr(v[1]))


def roundup(size):
    k = MINSIZE  # NOQA: F821 undefined name
    while k < size:
        k += k
    return k


def hitrate(loads, hits):
    if loads < 1:
        return 'n/a'
    return "%5.1f%%" % (100.0 * hits / loads)


def duration(secs):
    mm, ss = divmod(secs, 60)
    hh, mm = divmod(mm, 60)
    if hh:
        return "%d:%02d:%02d" % (hh, mm, ss)
    if mm:
        return "%d:%02d" % (mm, ss)
    return "%d" % ss


nre = re.compile(r'([=-]?)(\d+)([.]\d*)?').match


def addcommas(n):
    sign, s, d = nre(str(n)).group(1, 2, 3)
    if d == '.0':
        d = ''

    result = s[-3:]
    s = s[:-3]
    while s:
        result = s[-3:]+','+result
        s = s[:-3]

    return (sign or '') + result + (d or '')


def maybe(f, p=0.5):
    if random.random() < p:
        f()


if __name__ == "__main__":
    sys.exit(main())



from zope.testing import setupstack

from ZEO import _forker as forker

from .. import client
from .. import server


class SyncTests(setupstack.TestCase):

    def instrument(self):
        self.__ping_calls = 0

        server = getattr(forker, self.__name + '_server')

        [zs] = getattr(server.server, 'zeo_storages_by_storage_id')['1']
        orig_ping = getattr(zs, 'ping')

        def ping():
            self.__ping_calls += 1
            return orig_ping()

        setattr(zs, 'ping', ping)

    def test_server_sync(self):
        self.__name = 's%s' % id(self)
        addr, stop = server(name=self.__name)

        # By default the client sync method is a noop:
        c = client(addr, wait_timeout=2)
        self.instrument()
        c.sync()
        self.assertEqual(self.__ping_calls, 0)
        c.close()

        # But if we pass server_sync:
        c = client(addr, wait_timeout=2, server_sync=True)
        self.instrument()
        c.sync()
        self.assertEqual(self.__ping_calls, 1)
        c.close()

        stop()



"""Clients can pass credentials to a server.

This is an experimental feature to enable server authentication and
authorization.
"""
import unittest

from zope.testing import setupstack

import ZEO.StorageServer

from .threaded import threaded_server_tests


class ClientAuthTests(setupstack.TestCase):

    def setUp(self):
        self.setUpDirectory()
        self.__register = ZEO.StorageServer.ZEOStorage.register

    def tearDown(self):
        ZEO.StorageServer.ZEOStorage.register = self.__register

    def test_passing_credentials(self):

        # First, we'll temporarily swap the storage server register
        # method with one that let's is see credentials that were passed:

        creds_log = []

        def register(zs, storage_id, read_only, credentials=self):
            creds_log.append(credentials)
            return self.__register(zs, storage_id, read_only)

        ZEO.StorageServer.ZEOStorage.register = register

        # Now start an in process server
        addr, stop = ZEO.server()

        # If we connect, without providing credentials, then no
        # credentials will be passed to register:

        client = ZEO.client(addr)

        self.assertEqual(creds_log, [self])
        client.close()
        creds_log.pop()

        # Even if we pass credentials, they'll be ignored
        creds = dict(user='me', password='123')
        client = ZEO.client(addr, credentials=creds)
        self.assertEqual(creds_log, [self])
        client.close()

        stop()


def test_suite():
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(ClientAuthTests)
    suite.layer = threaded_server_tests
    return suite



#!/usr/bin/env python2.3

"""Parse the BLATHER logging generated by ZEO2.

An example of the log format is:
2002-04-15T13:05:29 BLATHER(-100) ZEO Server storea(3235680, [714], 235339406490168806) ('10.0.26.30', 45514)  # NOQA: E501 line too long
"""

import re
import time


rx_time = re.compile(r'(\d\d\d\d-\d\d-\d\d)T(\d\d:\d\d:\d\d)')


def parse_time(line):
    """Return the time portion of a zLOG line in seconds or None."""
    mo = rx_time.match(line)
    if mo is None:
        return None
    date, time_ = mo.group(1, 2)
    date_l = [int(elt) for elt in date.split('-')]
    time_l = [int(elt) for elt in time_.split(':')]
    return int(time.mktime(date_l + time_l + [0, 0, 0]))


rx_meth = re.compile(r"zrpc:\d+ calling (\w+)\((.*)")


def parse_method(line):
    pass


def parse_line(line):
    """Parse a log entry and return time, method info, and client."""
    t = parse_time(line)
    if t is None:
        return None, None
    mo = rx_meth.search(line)
    if mo is None:
        return None, None
    meth_name = mo.group(1)
    meth_args = mo.group(2).strip()
    if meth_args.endswith(')'):
        meth_args = meth_args[:-1]
    meth_args = [s.strip() for s in meth_args.split(",")]
    m = meth_name, tuple(meth_args)
    return t, m


class TStats:

    counter = 1

    def __init__(self):
        self.id = TStats.counter
        TStats.counter += 1

    fields = ("time", "vote", "done", "user", "path")
    fmt = "%-24s %5s %5s %-15s %s"
    hdr = fmt % fields

    def report(self):
        """Print a report about the transaction"""
        if hasattr(self, "vote"):
            d_vote = self.vote - self.begin
        else:
            d_vote = "*"
        if hasattr(self, "finish"):
            d_finish = self.finish - self.begin
        else:
            d_finish = "*"
        print(self.fmt % (time.ctime(self.begin), d_vote, d_finish,
                          self.user, self.url))


class TransactionParser:

    def __init__(self):
        self.txns = {}
        self.skipped = 0

    def parse(self, line):
        t, m = parse_line(line)
        if t is None:
            return
        name = m[0]
        meth = getattr(self, name, None)
        if meth is not None:
            meth(t, m[1])

    def tpc_begin(self, time, args):
        t = TStats()
        t.begin = time
        t.user = args[1]
        t.url = args[2]
        t.objects = []
        tid = eval(args[0])
        self.txns[tid] = t

    def get_txn(self, args):
        tid = eval(args[0])
        try:
            return self.txns[tid]
        except KeyError:
            print("uknown tid", repr(tid))
            return None

    def tpc_finish(self, time, args):
        t = self.get_txn(args)
        if t is None:
            return
        t.finish = time

    def vote(self, time, args):
        t = self.get_txn(args)
        if t is None:
            return
        t.vote = time

    def get_txns(self):
        L = [(t.id, t) for t in self.txns.values()]
        L.sort()
        return [t for (id, t) in L]


if __name__ == "__main__":
    import fileinput

    p = TransactionParser()
    i = 0
    for line in fileinput.input():
        i += 1
        try:
            p.parse(line)
        except:  # NOQA: E722 bare except
            print("line", i)
            raise
    print("Transaction: %d" % len(p.txns))
    print(TStats.hdr)
    for txn in p.get_txns():
        txn.report()



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
import concurrent.futures
import contextlib
import logging
import os
import socket
import threading
import time

import transaction
from ZODB.Connection import TransactionMetaData
from ZODB.DB import DB
from ZODB.POSException import ReadOnlyError
from ZODB.tests.MinPO import MinPO
from ZODB.tests.StorageTestBase import StorageTestBase
from ZODB.tests.StorageTestBase import zodb_pickle
from ZODB.tests.StorageTestBase import zodb_unpickle

from ZEO.asyncio.marshal import encode
from ZEO.ClientStorage import ClientStorage
from ZEO.Exceptions import ClientDisconnected
from ZEO.tests import forker

from . import testssl


logger = logging.getLogger('ZEO.tests.ConnectionTests')

ZERO = '\0'*8


class TestClientStorage(ClientStorage):

    # most operations will wait for reconnection with
    # ``wait_timeout`` (default: ``30``).
    # This is bad for our tests: they will eventually succeed, but
    # take a very long time. Use a timeout more adapted to tests.
    def __init__(*args, **kw):
        return ClientStorage.__init__(
                *args, **(kw.update(dict(wait_timeout=4)) or kw))

    test_connection = False

    connection_count_for_tests = 0

    def notify_connected(self, conn, info):
        ClientStorage.notify_connected(self, conn, info)
        self.connection_count_for_tests += 1
        self.verify_result = conn.verify_result


class DummyDB:
    def invalidate(self, *args, **kwargs):
        pass

    def invalidateCache(self):
        pass

    transform_record_data = untransform_record_data = lambda self, data: data


class CommonSetupTearDown(StorageTestBase):
    """Common boilerplate"""

    __super_setUp = StorageTestBase.setUp
    __super_tearDown = StorageTestBase.tearDown
    keep = 0
    invq = None
    timeout = None
    db_class = DummyDB

    def setUp(self, before=None):
        """Test setup for connection tests.

        This starts only one server; a test may start more servers by
        calling self._newAddr() and then self.startServer(index=i)
        for i in 1, 2, ...
        """
        self.__super_setUp()
        logger.info("setUp() %s", self.id())
        self.file = 'storage_conf'
        self._servers = []
        self.caches = []
        self.addr = [('127.0.0.1', 0)]
        self.startServer()

    def tearDown(self):
        """Try to cause the tests to halt"""

        if getattr(self, '_storage', None) is not None:
            self._storage.close()
            if hasattr(self._storage, 'cleanup'):
                logger.debug("cleanup storage %s" %
                             self._storage.__name__)
                self._storage.cleanup()
        for stop in self._servers:
            stop()

        for c in self.caches:
            for i in 0, 1:
                for ext in "", ".trace", ".lock":
                    path = f'{c}-1.zec{ext}'
                    # On Windows before 2.3, we don't have a way to wait for
                    # the spawned server(s) to close, and they inherited
                    # file descriptors for our open files.  So long as those
                    # processes are alive, we can't delete the files.  Try
                    # a few times then give up.
                    need_to_delete = False
                    if os.path.exists(path):
                        need_to_delete = True
                        for dummy in range(5):
                            try:
                                os.unlink(path)
                            except:  # NOQA: E722 bare except
                                time.sleep(0.5)
                            else:
                                need_to_delete = False
                                break
                    if need_to_delete:
                        os.unlink(path)  # sometimes this is just gonna fail
        self.__super_tearDown()

    def _newAddr(self):
        self.addr.append(self._getAddr())

    def _getAddr(self):
        return '127.0.0.1', forker.get_port(self)

    def getConfig(self, path, create, read_only):
        raise NotImplementedError

    cache_id = 1

    def openClientStorage(self, cache=None, cache_size=200000, wait=1,
                          read_only=0, read_only_fallback=0,
                          username=None, password=None, realm=None,
                          client_label=None):
        if cache is None:
            cache = str(self.__class__.cache_id)
            self.__class__.cache_id += 1
        self.caches.append(cache)
        storage = TestClientStorage(self.addr,
                                    client=cache,
                                    var='.',
                                    cache_size=cache_size,
                                    wait=wait,
                                    min_disconnect_poll=0.1,
                                    read_only=read_only,
                                    read_only_fallback=read_only_fallback,
                                    client_label=client_label,
                                    **self._client_options())
        storage.registerDB(DummyDB())
        return storage

    def _client_options(self):
        return {}

    def getServerConfig(self, addr, ro_svr):
        zconf = forker.ZEOConfig(addr, log='server.log')
        if ro_svr:
            zconf.read_only = 1
        if self.invq:
            zconf.invalidation_queue_size = self.invq
        if self.timeout:
            zconf.transaction_timeout = self.timeout
        return zconf

    def startServer(self, create=1, index=0, read_only=0, ro_svr=0, keep=None,
                    path=None, **kw):
        addr = self.addr[index]
        logger.info("startServer(create=%d, index=%d, read_only=%d) @ %s" %
                    (create, index, read_only, addr))
        if path is None:
            path = "%s.%d" % (self.file, index)
        sconf = self.getConfig(path, create, read_only)
        zconf = self.getServerConfig(addr, ro_svr)
        if keep is None:
            keep = self.keep
        zeoport, stop = forker.start_zeo_server(
            sconf, zconf, addr[1], keep, **kw)
        self._servers.append(stop)
        if addr[1] == 0:
            self.addr[index] = zeoport

    def shutdownServer(self, index=0):
        logger.info("shutdownServer(index=%d) @ %s" %
                    (index, self._servers[index]))
        stop = self._servers[index]
        if stop is not None:
            stop()
            self._servers[index] = lambda: None

    def pollUp(self, timeout=30.0, storage=None):
        if storage is None:
            storage = self._storage
        storage.server_status(timeout=timeout)

    def pollDown(self, timeout=30.0):
        # Poll until we're disconnected.
        now = time.time()
        giveup = now + timeout
        while self._storage.is_connected():
            now = time.time()
            if now > giveup:
                self.fail("timed out waiting for storage to disconnect")
            time.sleep(0.1)


class ConnectionTests(CommonSetupTearDown):
    """Tests that explicitly manage the server process.

    To test the cache or re-connection, these test cases explicit
    start and stop a ZEO storage server.
    """

    def checkMultipleAddresses(self):
        for i in range(4):
            self._newAddr()
        self._storage = self.openClientStorage('test', 100000)
        oid = self._storage.new_oid()
        obj = MinPO(12)
        self._dostore(oid, data=obj)
        self._storage.close()

    def checkReadOnlyClient(self):
        # Open a read-only client to a read-write server; stores fail

        # Start a read-only client for a read-write server
        self._storage = self.openClientStorage(read_only=1)
        # Stores should fail here
        self.assertRaises(ReadOnlyError, self._dostore)
        self._storage.close()

    def checkReadOnlyServer(self):
        # Open a read-only client to a read-only *server*; stores fail

        # We don't want the read-write server created by setUp()
        self.shutdownServer()
        self._servers = []
        # Start a read-only server
        self.startServer(create=0, index=0, ro_svr=1)
        # Start a read-only client
        self._storage = self.openClientStorage(read_only=1)
        # Stores should fail here
        self.assertRaises(ReadOnlyError, self._dostore)
        self._storage.close()
        # Get rid of the 'test left new threads behind' warning
        time.sleep(0.1)

    def checkReadOnlyFallbackWritable(self):
        # Open a fallback client to a read-write server; stores succeed

        # Start a read-only-fallback client for a read-write server
        self._storage = self.openClientStorage(read_only_fallback=1)
        # Stores should succeed here
        self._dostore()
        self._storage.close()

    def checkReadOnlyFallbackReadOnlyServer(self):
        # Open a fallback client to a read-only *server*; stores fail

        # We don't want the read-write server created by setUp()
        self.shutdownServer()
        self._servers = []
        # Start a read-only server
        self.startServer(create=0, index=0, ro_svr=1)
        # Start a read-only-fallback client
        self._storage = self.openClientStorage(read_only_fallback=1)
        self.assertTrue(self._storage.isReadOnly())
        # Stores should fail here
        self.assertRaises(ReadOnlyError, self._dostore)
        self._storage.close()

    def checkDisconnectionError(self):
        # Make sure we get a ClientDisconnected when we try to read an
        # object when we're not connected to a storage server and the
        # object is not in the cache.
        self.shutdownServer()
        self._storage = self.openClientStorage('test', 1000, wait=0)
        with short_timeout(self):
            self.assertRaises(ClientDisconnected,
                              self._storage.load, b'fredwash', '')
        self._storage.close()

    def checkBasicPersistence(self):
        # Verify cached data persists across client storage instances.

        # To verify that the cache is being used, the test closes the
        # server and then starts a new client with the server down.
        # When the server is down, a load() gets the data from its cache.

        self._storage = self.openClientStorage('test', 100000)
        oid = self._storage.new_oid()
        obj = MinPO(12)
        revid1 = self._dostore(oid, data=obj)
        self._storage.close()
        self.shutdownServer()
        self._storage = self.openClientStorage('test', 100000, wait=0)
        data, revid2 = self._storage.load(oid, '')
        self.assertEqual(zodb_unpickle(data), MinPO(12))
        self.assertEqual(revid1, revid2)
        self._storage.close()

    def checkDisconnectedCacheWorks(self):
        # Check that the cache works when the client is disconnected.
        self._storage = self.openClientStorage('test')
        oid1 = self._storage.new_oid()
        obj1 = MinPO("1" * 500)
        self._dostore(oid1, data=obj1)
        oid2 = self._storage.new_oid()
        obj2 = MinPO("2" * 500)
        self._dostore(oid2, data=obj2)
        expected1 = self._storage.load(oid1, '')
        expected2 = self._storage.load(oid2, '')

        # Shut it all down, and try loading from the persistent cache file
        # without a server present.
        self._storage.close()
        self.shutdownServer()
        self._storage = self.openClientStorage('test', wait=False)
        self.assertEqual(expected1, self._storage.load(oid1, ''))
        self.assertEqual(expected2, self._storage.load(oid2, ''))
        self._storage.close()

    def checkDisconnectedCacheFails(self):
        # Like checkDisconnectedCacheWorks above, except the cache
        # file is so small that only one object can be remembered.
        self._storage = self.openClientStorage('test', cache_size=900)
        oid1 = self._storage.new_oid()
        obj1 = MinPO("1" * 500)
        self._dostore(oid1, data=obj1)
        oid2 = self._storage.new_oid()
        obj2 = MinPO("2" * 500)
        # The cache file is so small that adding oid2 will evict oid1.
        self._dostore(oid2, data=obj2)
        expected2 = self._storage.load(oid2, '')

        # Shut it all down, and try loading from the persistent cache file
        # without a server present.
        self._storage.close()
        self.shutdownServer()
        self._storage = self.openClientStorage('test', cache_size=900,
                                               wait=False)
        # oid2 should still be in cache.
        self.assertEqual(expected2, self._storage.load(oid2, ''))
        # But oid1 should have been purged, so that trying to load it will
        # try to fetch it from the (non-existent) ZEO server.
        with short_timeout(self):
            self.assertRaises(ClientDisconnected, self._storage.load, oid1, '')
        self._storage.close()

    def checkVerificationInvalidationPersists(self):
        # This tests a subtle invalidation bug from ZODB 3.3:
        # invalidations processed as part of ZEO cache verification acted
        # kinda OK wrt the in-memory cache structures, but had no effect
        # on the cache file.  So opening the file cache again could
        # incorrectly believe that a previously invalidated object was
        # still current.  This takes some effort to set up.

        # First, using a persistent cache ('test'), create an object
        # MinPO(13).  We used to see this again at the end of this test,
        # despite that we modify it, and despite that it gets invalidated
        # in 'test', before the end.
        self._storage = self.openClientStorage('test')
        oid = self._storage.new_oid()
        obj = MinPO(13)
        self._dostore(oid, data=obj)
        self._storage.close()

        # Now modify obj via a temp connection.  `test` won't learn about
        # this until we open a connection using `test` again.
        self._storage = self.openClientStorage()
        pickle, rev = self._storage.load(oid, '')
        newobj = zodb_unpickle(pickle)
        self.assertEqual(newobj, obj)
        newobj.value = 42  # .value *should* be 42 forever after now, not 13
        self._dostore(oid, data=newobj, revid=rev)
        self._storage.close()

        # Open 'test' again.  `oid` in this cache should be (and is)
        # invalidated during cache verification.  The bug was that it
        # got invalidated (kinda) in memory, but not in the cache file.
        self._storage = self.openClientStorage('test')

        # The invalidation happened already.  Now create and store a new
        # object before closing this storage:  this is so `test` believes
        # it's seen transactions beyond the one that invalidated `oid`, so
        # that the *next* time we open `test` it doesn't process another
        # invalidation for `oid`.  It's also important that we not try to
        # load `oid` now:  because it's been (kinda) invalidated in the
        # cache's memory structures, loading it now would fetch the
        # current revision from the server, thus hiding the bug.
        obj2 = MinPO(666)
        oid2 = self._storage.new_oid()
        self._dostore(oid2, data=obj2)
        self._storage.close()

        # Finally, open `test` again and load `oid`.  `test` believes
        # it's beyond the transaction that modified `oid`, so its view
        # of whether it has an up-to-date `oid` comes solely from the disk
        # file, unaffected by cache verification.
        self._storage = self.openClientStorage('test')
        pickle, rev = self._storage.load(oid, '')
        newobj_copy = zodb_unpickle(pickle)
        # This used to fail, with
        #     AssertionError: MinPO(13) != MinPO(42)
        # That is, `test` retained a stale revision of the object on disk.
        self.assertEqual(newobj_copy, newobj)
        self._storage.close()

    def checkBadMessage1(self):
        # not even close to a real message
        self._bad_message(b"salty")

    def checkBadMessage2(self):
        # just like a real message, but with an unpicklable argument
        global Hack

        class Hack:
            pass

        msg = encode(1, 0, "foo", (Hack(),))
        self._bad_message(msg)
        del Hack

    def _bad_message(self, msg):
        # Establish a connection, then send the server an ill-formatted
        # request.  Verify that the connection is closed and that it is
        # possible to establish a new connection.

        self._storage = self.openClientStorage()
        self._dostore()

        generation = self._storage._connection_generation

        future = concurrent.futures.Future()

        def write():
            try:
                self._storage._server.client.protocol.write_message(msg)
            except Exception as exc:
                future.set_exception(exc)
            else:
                future.set_result(None)

        # break into the internals to send a bogus message
        self._storage._server.loop.call_soon_threadsafe(write)
        future.result()

        # If we manage to call _dostore before the server disconnects
        # us, we'll get a ClientDisconnected error.  When we retry, it
        # will succeed.  It will succeed because:
        # - _dostore calls tpc_abort
        # - tpc_abort makes a synchronous call to the server to abort
        #   the transaction
        # - when disconnected, synchronous calls are blocked for a little
        #   while while reconnecting (or they timeout of it takes too long).
        try:
            self._dostore()
        except ClientDisconnected:
            self._dostore()

        self.assertTrue(self._storage._connection_generation > generation)

    # Test case for multiple storages participating in a single
    # transaction.  This is not really a connection test, but it needs
    # about the same infrastructure (several storage servers).

    # TODO: with the current ZEO code, this occasionally fails.
    # That's the point of this test. :-)

    def NOcheckMultiStorageTransaction(self):
        # Configuration parameters (larger values mean more likely deadlocks)
        N = 2
        # These don't *have* to be all the same, but it's convenient this way
        self.nservers = N
        self.nthreads = N
        self.ntrans = N
        self.nobj = N

        # Start extra servers
        for i in range(1, self.nservers):
            self._newAddr()
            self.startServer(index=i)

        # Spawn threads that each do some transactions on all storages
        threads = []
        try:
            for i in range(self.nthreads):
                t = MSTThread(self, "T%d" % i)
                threads.append(t)
                t.start()
            # Wait for all threads to finish
            for t in threads:
                t.join(60)
                self.assertFalse(t.is_alive(), "%s didn't die" % t.getName())
        finally:
            for t in threads:
                t.closeclients()

    def checkCrossDBInvalidations(self):
        db1 = DB(self.openClientStorage())
        c1 = db1.open()
        r1 = c1.root()

        r1["a"] = MinPO("a")
        transaction.commit()
        self.assertEqual(r1._p_state, 0)  # up-to-date

        db2 = DB(self.openClientStorage())
        r2 = db2.open().root()

        self.assertEqual(r2["a"].value, "a")

        r2["b"] = MinPO("b")
        transaction.commit()

        # Make sure the invalidation is received in the other client.
        # We've had problems with this timing out on "slow" and/or "very
        # busy" machines, so we increase the sleep time on each trip, and
        # are willing to wait quite a long time.
        for i in range(20):
            c1.sync()
            if r1._p_state == -1:
                break
            time.sleep(i / 10.0)
        self.assertEqual(r1._p_state, -1)  # ghost

        r1.keys()  # unghostify
        self.assertEqual(r1._p_serial, r2._p_serial)
        self.assertEqual(r1["b"].value, "b")

        db2.close()
        db1.close()

    def checkCheckForOutOfDateServer(self):
        # We don't want to connect a client to a server if the client
        # has seen newer transactions.
        self._storage = self.openClientStorage()
        self._dostore()
        self.shutdownServer()
        with short_timeout(self):
            self.assertRaises(ClientDisconnected,
                              self._storage.load, b'\0'*8, '')

        self.startServer()

        # No matter how long we wait, the client won't reconnect:
        time.sleep(2)
        with short_timeout(self):
            self.assertRaises(ClientDisconnected,
                              self._storage.load, b'\0'*8, '')


class SSLConnectionTests(ConnectionTests):

    def getServerConfig(self, addr, ro_svr):
        return testssl.server_config.replace(
            '127.0.0.1:0',
            '{}: {}\nread-only {}'.format(
                addr[0], addr[1], 'true' if ro_svr else 'false'))

    def _client_options(self):
        return {'ssl': testssl.client_ssl()}


class InvqTests(CommonSetupTearDown):
    invq = 3

    def checkQuickVerificationWith2Clients(self):
        perstorage = self.openClientStorage(cache="test", cache_size=4000)

        self._storage = self.openClientStorage()
        oid = self._storage.new_oid()
        oid2 = self._storage.new_oid()
        # When we create a new storage, it should always do a full
        # verification
        self.assertEqual(self._storage.verify_result, "empty cache")
        # do two storages of the object to make sure an invalidation
        # message is generated
        revid = self._dostore(oid)
        revid = self._dostore(oid, revid)
        # Create a second object and revision to guarantee it doesn't
        # show up in the list of invalidations sent when perstore restarts.
        revid2 = self._dostore(oid2)
        revid2 = self._dostore(oid2, revid2)

        forker.wait_until(
            lambda:
            perstorage.lastTransaction() == self._storage.lastTransaction())

        perstorage.load(oid, '')
        perstorage.close()

        forker.wait_until(lambda: os.path.exists('test-1.zec'))

        revid = self._dostore(oid, revid)

        perstorage = self.openClientStorage(cache="test")
        self.assertEqual(perstorage.verify_result, "quick verification")

        self.assertEqual(perstorage.load(oid, ''),
                         self._storage.load(oid, ''))
        perstorage.close()

    def checkVerificationWith2ClientsInvqOverflow(self):
        perstorage = self.openClientStorage(cache="test")
        self.assertEqual(perstorage.verify_result, "empty cache")

        self._storage = self.openClientStorage()
        oid = self._storage.new_oid()
        # When we create a new storage, it should always do a full
        # verification
        self.assertEqual(self._storage.verify_result, "empty cache")
        # do two storages of the object to make sure an invalidation
        # message is generated
        revid = self._dostore(oid)
        revid = self._dostore(oid, revid)
        forker.wait_until(
            "Client has seen all of the transactions from the server",
            lambda:
            perstorage.lastTransaction() == self._storage.lastTransaction()
            )
        perstorage.load(oid, '')
        perstorage.close()

        # the test code sets invq bound to 2
        for i in range(5):
            revid = self._dostore(oid, revid)

        perstorage = self.openClientStorage(cache="test")
        self.assertEqual(perstorage.verify_result, "cache too old, clearing")
        self.assertEqual(self._storage.load(oid, '')[1], revid)
        self.assertEqual(perstorage.load(oid, ''),
                         self._storage.load(oid, ''))

        perstorage.close()


class ReconnectionTests(CommonSetupTearDown):
    # The setUp() starts a server automatically.  In order for its
    # state to persist, we set the class variable keep to 1.  In
    # order for its state to be cleaned up, the last startServer()
    # call in the test must pass keep=0.
    keep = 1
    invq = 2

    def checkReadOnlyStorage(self):
        # Open a read-only client to a read-only *storage*; stores fail

        # We don't want the read-write server created by setUp()
        self.shutdownServer()
        self._servers = []
        # Start a read-only server
        self.startServer(create=0, index=0, read_only=1, keep=0)
        # Start a read-only client
        self._storage = self.openClientStorage(read_only=1)
        # Stores should fail here
        self.assertRaises(ReadOnlyError, self._dostore)

    def checkReadOnlyFallbackReadOnlyStorage(self):
        # Open a fallback client to a read-only *storage*; stores fail

        # We don't want the read-write server created by setUp()
        self.shutdownServer()
        self._servers = []
        # Start a read-only server
        self.startServer(create=0, index=0, read_only=1, keep=0)
        # Start a read-only-fallback client
        self._storage = self.openClientStorage(read_only_fallback=1)
        # Stores should fail here
        self.assertRaises(ReadOnlyError, self._dostore)

    # TODO:  Compare checkReconnectXXX() here to checkReconnection()
    # further down.  Is the code here hopelessly naive, or is
    # checkReconnection() overwrought?

    def checkReconnectWritable(self):
        # A read-write client reconnects to a read-write server

        # Start a client
        self._storage = self.openClientStorage()
        # Stores should succeed here
        self._dostore()

        # Shut down the server
        self.shutdownServer()
        self._servers = []
        # Poll until the client disconnects
        self.pollDown()
        # Stores should fail now
        with short_timeout(self):
            self.assertRaises(ClientDisconnected, self._dostore)

        # Restart the server
        self.startServer(create=0)
        # Poll until the client connects
        self.pollUp()
        # Stores should succeed here
        self._dostore()
        self._storage.close()

    def checkReconnectReadOnly(self):
        # A read-only client reconnects from a read-write to a
        # read-only server

        # Start a client
        self._storage = self.openClientStorage(read_only=1)
        # Stores should fail here
        self.assertRaises(ReadOnlyError, self._dostore)

        # Shut down the server
        self.shutdownServer()
        self._servers = []
        # Poll until the client disconnects
        self.pollDown()
        # Stores should still fail
        self.assertRaises(ReadOnlyError, self._dostore)

        # Restart the server
        self.startServer(create=0, read_only=1, keep=0)
        # Poll until the client connects
        self.pollUp()
        # Stores should still fail
        self.assertRaises(ReadOnlyError, self._dostore)

    def checkReconnectFallback(self):
        # A fallback client reconnects from a read-write to a
        # read-only server

        # Start a client in fallback mode
        self._storage = self.openClientStorage(read_only_fallback=1)
        # Stores should succeed here
        self._dostore()

        # Shut down the server
        self.shutdownServer()
        self._servers = []
        # Poll until the client disconnects
        self.pollDown()
        # Stores should fail now
        with short_timeout(self):
            self.assertRaises(ClientDisconnected, self._dostore)

        # Restart the server
        self.startServer(create=0, read_only=1, keep=0)
        # Poll until the client connects
        self.pollUp()
        # Stores should fail here
        self.assertRaises(ReadOnlyError, self._dostore)

    def checkReconnectUpgrade(self):
        # A fallback client reconnects from a read-only to a
        # read-write server

        # We don't want the read-write server created by setUp()
        self.shutdownServer()
        self._servers = []
        # Start a read-only server
        self.startServer(create=0, read_only=1)
        # Start a client in fallback mode
        self._storage = self.openClientStorage(read_only_fallback=1)
        # Stores should fail here
        self.assertRaises(ReadOnlyError, self._dostore)

        # Shut down the server
        self.shutdownServer()
        self._servers = []
        # Poll until the client disconnects
        self.pollDown()

        # Accesses should fail now
        with short_timeout(self):
            self.assertRaises(ClientDisconnected, self._storage.ping)

        # Restart the server, this time read-write
        self.startServer(create=0, keep=0)
        # Poll until the client sconnects
        self.pollUp()
        # Stores should now succeed
        self._dostore()

    def checkReconnectSwitch(self):
        # A fallback client initially connects to a read-only server,
        # then discovers a read-write server and switches to that

        # We don't want the read-write server created by setUp()
        self.shutdownServer()
        self._servers = []
        # Allocate a second address (for the second server)
        self._newAddr()

        # Start a read-only server
        self.startServer(create=0, index=0, read_only=1, keep=0)
        # Start a client in fallback mode
        self._storage = self.openClientStorage(read_only_fallback=1)
        # Stores should fail here
        self.assertRaises(ReadOnlyError, self._dostore)

        # Start a read-write server
        self.startServer(index=1, read_only=0, keep=0)
        # After a while, stores should work
        for i in range(300):  # Try for 30 seconds
            try:
                self._dostore()
                break
            except (ClientDisconnected, ReadOnlyError):
                # If the client isn't connected at all, sync() returns
                # quickly and the test fails because it doesn't wait
                # long enough for the client.
                time.sleep(0.1)
        else:
            self.fail("Couldn't store after starting a read-write server")

    def checkNoVerificationOnServerRestart(self):
        self._storage = self.openClientStorage()
        # When we create a new storage, it should always do a full
        # verification
        self.assertEqual(self._storage.verify_result, "empty cache")
        self._dostore()
        self.shutdownServer()
        self.pollDown()
        self._storage.verify_result = None
        self.startServer(create=0, keep=0)
        self.pollUp()
        # There were no transactions committed, so no verification
        # should be needed.
        self.assertEqual(self._storage.verify_result, "Cache up to date")

    def checkNoVerificationOnServerRestartWith2Clients(self):
        perstorage = self.openClientStorage(cache="test")
        self.assertEqual(perstorage.verify_result, "empty cache")

        self._storage = self.openClientStorage()
        oid = self._storage.new_oid()
        # When we create a new storage, it should always do a full
        # verification
        self.assertEqual(self._storage.verify_result, "empty cache")
        # do two storages of the object to make sure an invalidation
        # message is generated
        revid = self._dostore(oid)
        revid = self._dostore(oid, revid)
        forker.wait_until(
            "Client has seen all of the transactions from the server",
            lambda:
            perstorage.lastTransaction() == self._storage.lastTransaction()
            )
        perstorage.load(oid, '')
        self.shutdownServer()
        self.pollDown()
        self._storage.verify_result = None
        perstorage.verify_result = None
        logger.info('2ALLBEEF')
        self.startServer(create=0, keep=0)
        self.pollUp()
        self.pollUp(storage=perstorage)
        # There were no transactions committed, so no verification
        # should be needed.
        self.assertEqual(self._storage.verify_result, "Cache up to date")
        self.assertEqual(perstorage.verify_result, "Cache up to date")
        perstorage.close()
        self._storage.close()

    def checkDisconnectedAbort(self):
        self._storage = self.openClientStorage()
        self._dostore()
        oids = [self._storage.new_oid() for i in range(5)]
        txn = TransactionMetaData()
        self._storage.tpc_begin(txn)
        for oid in oids:
            data = zodb_pickle(MinPO(oid))
            self._storage.store(oid, None, data, '', txn)
        self.shutdownServer()
        with short_timeout(self):
            self.assertRaises(ClientDisconnected, self._storage.tpc_vote, txn)
        self.startServer(create=0)
        # if we do not wait for reconnection,
        # small timing variations in the reconnection process
        # can cause occasional ``ClientDisconnected`` exceptions
        # making the test prone to race conditions.
        # Of course, in real life the followup operations
        # could happen while reconnecting but ``ClientDisconnected``
        # would be a good in those cases
        self.pollUp(2)  # await reconnection
        self._storage.tpc_abort(txn)
        self._dostore()  # stores in a new transaction

        # This test is supposed to cover the following error, although
        # I don't have much confidence that it does.  The likely
        # explanation for the error is that the _tbuf contained
        # objects that weren't in the _seriald, because the client was
        # interrupted waiting for tpc_vote() to return.  When the next
        # transaction committed, it tried to do something with the
        # bogus _tbuf entries.  The explanation is wrong/incomplete,
        # because tpc_begin() should clear the _tbuf.

        # 2003-01-15T15:44:19 ERROR(200) ZODB A storage error occurred
        # in the last phase of a two-phase commit.  This shouldn't happen.

        # Traceback (innermost last):
        # Module ZODB.Transaction, line 359, in _finish_one
        # Module ZODB.Connection, line 691, in tpc_finish
        # Module ZEO.ClientStorage, line 679, in tpc_finish
        # Module ZEO.ClientStorage, line 709, in _update_cache
        # KeyError: ...

    def checkReconnection(self):
        # Check that the client reconnects when a server restarts.

        self._storage = self.openClientStorage()
        oid = self._storage.new_oid()
        obj = MinPO(12)
        self._dostore(oid, data=obj)
        logger.info("checkReconnection(): About to shutdown server")
        self.shutdownServer()
        logger.info("checkReconnection(): About to restart server")
        self.startServer(create=0)
        forker.wait_until('reconnect', self._storage.is_connected)
        oid = self._storage.new_oid()
        obj = MinPO(12)
        while 1:
            try:
                self._dostore(oid, data=obj)
                break
            except ClientDisconnected:
                # Maybe the exception mess is better now
                logger.info("checkReconnection(): Error after"
                            " server restart; retrying.", exc_info=True)
                transaction.abort()
            # Give the other thread a chance to run.
            time.sleep(0.1)
        logger.info("checkReconnection(): finished")
        self._storage.close()

    def checkMultipleServers(self):
        # Crude test-- just start two servers and do a commit at each one.

        self._newAddr()
        self._storage = self.openClientStorage('test', 100000)
        self._dostore()
        self.shutdownServer(index=0)

        # When we start the second server, we use file data file from
        # the original server so tha the new server is a replica of
        # the original.  We need this because ClientStorage won't use
        # a server if the server's last transaction is earlier than
        # what the client has seen.
        self.startServer(index=1, path=self.file+'.0', create=False)

        # If we can still store after shutting down one of the
        # servers, we must be reconnecting to the other server.

        did_a_store = 0
        for i in range(10):
            try:
                self._dostore()
                did_a_store = 1
                break
            except ClientDisconnected:
                time.sleep(0.5)
        self.assertTrue(did_a_store)
        self._storage.close()


class TimeoutTests(CommonSetupTearDown):
    timeout = 1

    def checkTimeout(self):
        self._storage = storage = self.openClientStorage()
        txn = TransactionMetaData()
        storage.tpc_begin(txn)
        storage.tpc_vote(txn)
        time.sleep(2)
        with short_timeout(self):
            self.assertRaises(ClientDisconnected, storage.tpc_finish, txn)

        # Make sure it's logged as CRITICAL
        with open("server.log") as f:
            for line in f:
                if ('Transaction timeout after' in line) and \
                   ('CRITICAL ZEO.StorageServer' in line):
                    break
            else:
                self.fail('bad logging')

        storage.tpc_abort(txn)
        storage.close()

    def checkTimeoutOnAbort(self):
        storage = self.openClientStorage()
        txn = TransactionMetaData()
        storage.tpc_begin(txn)
        storage.tpc_vote(txn)
        storage.tpc_abort(txn)
        storage.close()

    def checkTimeoutOnAbortNoLock(self):
        storage = self.openClientStorage()
        txn = TransactionMetaData()
        storage.tpc_begin(txn)
        storage.tpc_abort(txn)
        storage.close()

    def checkTimeoutAfterVote(self):
        self._storage = storage = self.openClientStorage()
        # Assert that the zeo cache is empty
        self.assertTrue(not list(storage._cache.contents()))
        # Create the object
        oid = storage.new_oid()
        obj = MinPO(7)
        # Now do a store, sleeping before the finish so as to cause a timeout
        t = TransactionMetaData()
        old_connection_count = storage.connection_count_for_tests
        storage.tpc_begin(t)
        storage.store(oid, ZERO, zodb_pickle(obj), '', t)
        storage.tpc_vote(t)
        # Now sleep long enough for the storage to time out
        time.sleep(3)
        self.assertTrue(
            (not storage.is_connected())
            or
            (storage.connection_count_for_tests > old_connection_count)
            )
        storage._wait()
        self.assertTrue(storage.is_connected())
        # We expect finish to fail
        self.assertRaises(ClientDisconnected, storage.tpc_finish, t)
        # The cache should still be empty
        self.assertTrue(not list(storage._cache.contents()))
        # Load should fail since the object should not be in either the cache
        # or the server.
        self.assertRaises(KeyError, storage.load, oid, '')
        storage.tpc_abort(t)


class MSTThread(threading.Thread):

    __super_init = threading.Thread.__init__

    def __init__(self, testcase, name):
        self.__super_init(name=name)
        self.testcase = testcase
        self.clients = []

    def run(self):
        tname = self.getName()
        testcase = self.testcase

        # Create client connections to each server
        clients = self.clients
        for i in range(len(testcase.addr)):
            c = testcase.openClientStorage(addr=testcase.addr[i])
            c.__name = "C%d" % i
            clients.append(c)

        for i in range(testcase.ntrans):
            # Because we want a transaction spanning all storages,
            # we can't use _dostore().  This is several _dostore() calls
            # expanded in-line (mostly).

            # Create oid->serial mappings
            for c in clients:
                c.__oids = []
                c.__serials = {}

            # Begin a transaction
            t = TransactionMetaData()
            for c in clients:
                # print("%s.%s.%s begin" % (tname, c.__name, i))
                c.tpc_begin(t)

            for j in range(testcase.nobj):
                for c in clients:
                    # Create and store a new object on each server
                    oid = c.new_oid()
                    c.__oids.append(oid)
                    data = MinPO("%s.%s.t%d.o%d" % (tname, c.__name, i, j))
                    # print(data.value)
                    data = zodb_pickle(data)
                    c.store(oid, ZERO, data, '', t)

            # Vote on all servers and handle serials
            for c in clients:
                # print("%s.%s.%s vote" % (tname, c.__name, i))
                c.tpc_vote(t)

            # Finish on all servers
            for c in clients:
                # print("%s.%s.%s finish\n" % (tname, c.__name, i))
                c.tpc_finish(t)

            for c in clients:
                # Check that we got serials for all oids
                for oid in c.__oids:
                    testcase.assertIn(oid, c.__serials)
                # Check that we got serials for no other oids
                for oid in c.__serials.keys():
                    testcase.assertIn(oid, c.__oids)

    def closeclients(self):
        # Close clients opened by run()
        for c in self.clients:
            try:
                c.close()
            except:  # NOQA: E722 bare except
                pass


@contextlib.contextmanager
def short_timeout(self):
    old = self._storage._server.timeout
    self._storage._server.timeout = 1
    yield
    self._storage._server.timeout = old


# Run IPv6 tests if V6 sockets are supported
try:
    with socket.socket(socket.AF_INET6, socket.SOCK_STREAM) as s:
        pass
except (OSError, AttributeError):
    pass
else:
    class V6Setup:
        def _getAddr(self):
            return '::1', forker.get_port(self)

    _g = globals()
    for name, value in tuple(_g.items()):
        if isinstance(value, type) and issubclass(value, CommonSetupTearDown):
            _g[name+"V6"] = type(name+"V6", (V6Setup, value), {})



import unittest

from ZEO.asyncio.marshal import encode
from ZEO.asyncio.marshal import pickle_server_decode


try:
    from ZopeUndo.Prefix import Prefix
except ImportError:
    _HAVE_ZOPE_UNDO = False
else:
    _HAVE_ZOPE_UNDO = True


class MarshalTests(unittest.TestCase):

    @unittest.skipUnless(_HAVE_ZOPE_UNDO, 'ZopeUndo is not installed')
    def testServerDecodeZopeUndoFilter(self):
        # this is an example (1) of Zope2's arguments for
        # undoInfo call. Arguments are encoded by ZEO client
        # and decoded by server. The operation must be idempotent.
        # (1) https://github.com/zopefoundation/Zope/blob/2.13/src/App/Undo.py#L111  # NOQA: E501 line too long
        args = (0, 20, {'user_name': Prefix('test')})
        # test against repr because Prefix __eq__ operator
        # doesn't compare Prefix with Prefix but only
        # Prefix with strings. see Prefix.__doc__
        self.assertEqual(
            repr(pickle_server_decode(encode(*args))),
            repr(args)
        )



import os
import ssl
import unittest
from unittest import mock

from ZODB.config import storageFromString

from .. import runzeo
from ..Exceptions import ClientDisconnected
from .testConfig import ZEOConfigTestBase
from .threaded import threaded_server_tests


here = os.path.dirname(__file__)
server_cert = os.path.join(here, 'server.pem')
server_key = os.path.join(here, 'server_key.pem')
serverpw_cert = os.path.join(here, 'serverpw.pem')
serverpw_key = os.path.join(here, 'serverpw_key.pem')
client_cert = os.path.join(here, 'client.pem')
client_key = os.path.join(here, 'client_key.pem')


class SSLConfigTest(ZEOConfigTestBase):

    def test_ssl_basic(self):
        # This shows that configuring ssl has an actual effect on connections.
        # Other SSL configuration tests will be Mockiavellian.

        # Also test that an SSL connection mismatch doesn't kill
        # the server loop.

        # An SSL client can't talk to a non-SSL server:
        addr, stop = self.start_server()
        with self.assertRaises(ClientDisconnected):
            self.start_client(
                addr,
                f"""<ssl>
                certificate {client_cert}
                key {client_key}
                </ssl>""",
                wait_timeout=1)

        # But a non-ssl one can:
        client = self.start_client(addr)
        self._client_assertions(client, addr)
        client.close()
        stop()

        # A non-SSL client can't talk to an SSL server:
        addr, stop = self.start_server(
            f"""<ssl>
            certificate {server_cert}
            key {server_key}
            authenticate {client_cert}
            </ssl>"""
            )
        with self.assertRaises(ClientDisconnected):
            self.start_client(addr, wait_timeout=1)

        # But an SSL one can:
        client = self.start_client(
            addr,
            f"""<ssl>
                certificate {client_cert}
                key {client_key}
                authenticate {server_cert}
                server-hostname zodb.org
                </ssl>""")
        self._client_assertions(client, addr)
        client.close()
        stop()

    def test_ssl_hostname_check(self):
        addr, stop = self.start_server(
            f"""<ssl>
            certificate {server_cert}
            key {server_key}
            authenticate {client_cert}
            </ssl>"""
            )

        # Connext with bad hostname fails:

        with self.assertRaises(ClientDisconnected):
            client = self.start_client(
                addr,
                f"""<ssl>
                    certificate {client_cert}
                    key {client_key}
                    authenticate {server_cert}
                    server-hostname example.org
                    </ssl>""",
                wait_timeout=1)

        # Connext with good hostname succeeds:
        client = self.start_client(
            addr,
            f"""<ssl>
                certificate {client_cert}
                key {client_key}
                authenticate {server_cert}
                server-hostname zodb.org
                </ssl>""")
        self._client_assertions(client, addr)
        client.close()
        stop()

    def test_ssl_pw(self):
        addr, stop = self.start_server(
            f"""<ssl>
            certificate {serverpw_cert}
            key {serverpw_key}
            authenticate {client_cert}
            password-function ZEO.tests.testssl.pwfunc
            </ssl>"""
            )
        stop()


@mock.patch('asyncio.ensure_future')
@mock.patch('asyncio.set_event_loop')
@mock.patch('asyncio.new_event_loop')
@mock.patch('ZEO.asyncio.client.new_event_loop')
@mock.patch('ZEO.asyncio.server.new_event_loop')
class SSLConfigTestMockiavellian(ZEOConfigTestBase):

    @mock.patch('ssl.create_default_context')
    def test_ssl_mockiavellian_server_no_ssl(self, factory, *_):
        server = create_server()
        self.assertFalse(factory.called)
        self.assertEqual(server.acceptor.ssl_context, None)
        server.close()

    def assert_context(
            self,
            server,
            factory, context,
            cert=(server_cert, server_key, None),
            verify_mode=ssl.CERT_REQUIRED,
            check_hostname=False,
            cafile=None, capath=None):
        factory.assert_called_with(
            ssl.Purpose.CLIENT_AUTH if server else ssl.Purpose.SERVER_AUTH,
            cafile=cafile, capath=capath)
        context.load_cert_chain.assert_called_with(*cert)
        self.assertEqual(context, factory.return_value)
        self.assertEqual(context.verify_mode, verify_mode)
        self.assertEqual(context.check_hostname, check_hostname)

    @mock.patch('ssl.create_default_context')
    def test_ssl_mockiavellian_server_ssl_no_auth(self, factory, *_):
        with self.assertRaises(SystemExit):
            # auth is required
            create_server(certificate=server_cert, key=server_key)

    @mock.patch('ssl.create_default_context')
    def test_ssl_mockiavellian_server_ssl_auth_file(self, factory, *_):
        server = create_server(
            certificate=server_cert, key=server_key, authenticate=__file__)
        context = server.acceptor.ssl_context
        self.assert_context(True, factory, context, cafile=__file__)
        server.close()

    @mock.patch('ssl.create_default_context')
    def test_ssl_mockiavellian_server_ssl_auth_dir(self, factory, *_):
        server = create_server(
            certificate=server_cert, key=server_key, authenticate=here)
        context = server.acceptor.ssl_context
        self.assert_context(True, factory, context, capath=here)
        server.close()

    @mock.patch('ssl.create_default_context')
    def test_ssl_mockiavellian_server_ssl_pw(self, factory, *_):
        server = create_server(
            certificate=server_cert,
            key=server_key,
            password_function='ZEO.tests.testssl.pwfunc',
            authenticate=here,
            )
        context = server.acceptor.ssl_context
        self.assert_context(True,
                            factory,
                            context,
                            (server_cert, server_key, pwfunc),
                            capath=here)
        server.close()

    @mock.patch('ssl.create_default_context')
    @mock.patch('ZEO.ClientStorage.ClientStorage')
    def test_ssl_mockiavellian_client_no_ssl(self, ClientStorage, factory, *_):
        ssl_client()
        self.assertFalse('ssl' in ClientStorage.call_args[1])
        self.assertFalse('ssl_server_hostname' in ClientStorage.call_args[1])

    @mock.patch('ssl.create_default_context')
    @mock.patch('ZEO.ClientStorage.ClientStorage')
    def test_ssl_mockiavellian_client_server_signed(
            self, ClientStorage, factory, *_):
        ssl_client(certificate=client_cert, key=client_key)
        context = ClientStorage.call_args[1]['ssl']
        self.assertEqual(ClientStorage.call_args[1]['ssl_server_hostname'],
                         None)
        self.assert_context(False,
                            factory,
                            context,
                            (client_cert, client_key, None),
                            check_hostname=True)

        context.load_default_certs.assert_called_with()

    @mock.patch('ssl.create_default_context')
    @mock.patch('ZEO.ClientStorage.ClientStorage')
    def test_ssl_mockiavellian_client_auth_dir(
            self, ClientStorage, factory, *_):
        ssl_client(
            certificate=client_cert, key=client_key, authenticate=here)
        context = ClientStorage.call_args[1]['ssl']
        self.assertEqual(ClientStorage.call_args[1]['ssl_server_hostname'],
                         None)
        self.assert_context(False,
                            factory,
                            context,
                            (client_cert, client_key, None),
                            capath=here,
                            check_hostname=True)
        context.load_default_certs.assert_not_called()

    @mock.patch('ssl.create_default_context')
    @mock.patch('ZEO.ClientStorage.ClientStorage')
    def test_ssl_mockiavellian_client_auth_file(
            self, ClientStorage, factory, *_):
        ssl_client(
            certificate=client_cert, key=client_key, authenticate=server_cert)
        context = ClientStorage.call_args[1]['ssl']
        self.assertEqual(ClientStorage.call_args[1]['ssl_server_hostname'],
                         None)
        self.assert_context(False,
                            factory,
                            context,
                            (client_cert, client_key, None),
                            cafile=server_cert,
                            check_hostname=True)
        context.load_default_certs.assert_not_called()

    @mock.patch('ssl.create_default_context')
    @mock.patch('ZEO.ClientStorage.ClientStorage')
    def test_ssl_mockiavellian_client_pw(
            self, ClientStorage, factory, *_):
        ssl_client(
            certificate=client_cert, key=client_key,
            password_function='ZEO.tests.testssl.pwfunc',
            authenticate=server_cert)
        context = ClientStorage.call_args[1]['ssl']
        self.assertEqual(ClientStorage.call_args[1]['ssl_server_hostname'],
                         None)
        self.assert_context(False,
                            factory,
                            context,
                            (client_cert, client_key, pwfunc),
                            cafile=server_cert,
                            check_hostname=True)

    @mock.patch('ssl.create_default_context')
    @mock.patch('ZEO.ClientStorage.ClientStorage')
    def test_ssl_mockiavellian_client_server_hostname(
            self, ClientStorage, factory, *_):
        ssl_client(
            certificate=client_cert, key=client_key, authenticate=server_cert,
            server_hostname='example.com')
        context = ClientStorage.call_args[1]['ssl']
        self.assertEqual(ClientStorage.call_args[1]['ssl_server_hostname'],
                         'example.com')
        self.assert_context(False,
                            factory,
                            context,
                            (client_cert, client_key, None),
                            cafile=server_cert,
                            check_hostname=True)

    @mock.patch('ssl.create_default_context')
    @mock.patch('ZEO.ClientStorage.ClientStorage')
    def test_ssl_mockiavellian_client_check_hostname(
            self, ClientStorage, factory, *_):
        ssl_client(
            certificate=client_cert, key=client_key, authenticate=server_cert,
            check_hostname=False)
        context = ClientStorage.call_args[1]['ssl']
        self.assertEqual(ClientStorage.call_args[1]['ssl_server_hostname'],
                         None)
        self.assert_context(False,
                            factory,
                            context,
                            (client_cert, client_key, None),
                            cafile=server_cert,
                            check_hostname=False)


def args(*a, **kw):
    return a, kw


def ssl_conf(**ssl_settings):
    if ssl_settings:
        ssl_conf = '<ssl>\n' + '\n'.join(
            '{} {}'.format(name.replace('_', '-'), value)
            for name, value in ssl_settings.items()
            ) + '\n</ssl>\n'
    else:
        ssl_conf = ''

    return ssl_conf


def ssl_client(**ssl_settings):
    return storageFromString(
        """%import ZEO

        <clientstorage>
          server 127.0.0.1:0
          {}
        </clientstorage>
        """.format(ssl_conf(**ssl_settings))
        )


def create_server(**ssl_settings):
    with open('conf', 'w') as f:
        f.write(
            """
            <zeo>
              address 127.0.0.1:0
              {}
            </zeo>
            <mappingstorage>
            </mappingstorage>
            """.format(ssl_conf(**ssl_settings)))

    options = runzeo.ZEOOptions()
    options.realize(['-C', 'conf'])
    s = runzeo.ZEOServer(options)
    s.open_storages()
    s.create_server()
    return s.server


def pwfunc():
    return '1234'


def test_suite():
    suite = unittest.TestSuite((
        unittest.defaultTestLoader.loadTestsFromTestCase(SSLConfigTest),
        unittest.defaultTestLoader.loadTestsFromTestCase(
            SSLConfigTestMockiavellian),
        ))
    suite.layer = threaded_server_tests
    return suite


# Helpers for other tests:
server_config = f"""
    <zeo>
      address 127.0.0.1:0
      <ssl>
        certificate {server_cert}
        key {server_key}
        authenticate {client_cert}
      </ssl>
    </zeo>
    """


def client_ssl(cafile=server_key,
               client_cert=client_cert,
               client_key=client_key,
               ):
    context = ssl.create_default_context(
        ssl.Purpose.SERVER_AUTH, cafile=server_cert)

    context.load_cert_chain(client_cert, client_key)
    context.verify_mode = ssl.CERT_REQUIRED
    context.check_hostname = False
    return context

# See
# https://discuss.pivotal.io/hc/en-us/articles/202653388-How-to-renew-an-expired-Apache-Web-Server-self-signed-certificate-using-the-OpenSSL-tool  # NOQA: E501
# for instructions on updating the server.pem (the certificate) if
# needed. server.pem.csr is the request.
# This should do it:
# openssl x509 -req -days 999999 -in src/ZEO/tests/server.pem.csr -signkey src/ZEO/tests/server_key.pem -out src/ZEO/tests/server.pem  # NOQA: E501
# If you need to create a new key first:
# openssl genrsa -out server_key.pem 2048
# These two files should then be copied to client_key.pem and client.pem.



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Support for marshaling ZEO messages

Not to be confused with marshaling objects in ZODB.

We currently use pickle. In the future, we may use a
Python-independent format, or possibly a minimal pickle subset.
"""

import logging
from io import BytesIO

from zodbpickle.pickle import Pickler

from .._compat import Unpickler
from ..shortrepr import short_repr


logger = logging.getLogger(__name__)


def encoder(protocol, server=False):
    """Return a non-thread-safe encoder
    """

    if protocol[:1] == b'M':
        from msgpack import packb

        default = server_default if server else None

        def encode(*args):
            return packb(
                args, use_bin_type=True, default=default)

        return encode
    else:
        assert protocol[:1] == b'Z'

    f = BytesIO()
    getvalue = f.getvalue
    seek = f.seek
    truncate = f.truncate
    pickler = Pickler(f, 3)
    pickler.fast = 1
    dump = pickler.dump

    def encode(*args):
        seek(0)
        truncate()
        dump(args)
        return getvalue()

    return encode


def encode(*args):

    return encoder(b'Z')(*args)


def decoder(protocol):
    if protocol[:1] == b'M':
        from msgpack import unpackb

        def msgpack_decode(data):
            """Decodes msg and returns its parts"""
            return unpackb(data, raw=False, use_list=False)

        return msgpack_decode
    else:
        assert protocol[:1] == b'Z'
        return pickle_decode


def pickle_decode(msg):
    """Decodes msg and returns its parts"""
    unpickler = Unpickler(BytesIO(msg))
    unpickler.find_global = find_global
    try:
        # PyPy, zodbpickle, the non-c-accelerated version
        unpickler.find_class = find_global
    except AttributeError:
        pass
    try:
        return unpickler.load()  # msgid, flags, name, args
    except:  # NOQA: E722 bare except
        logger.error("can't decode message: %s" % short_repr(msg))
        raise


def server_decoder(protocol):
    if protocol[:1] == b'M':
        return decoder(protocol)
    else:
        assert protocol[:1] == b'Z'
        return pickle_server_decode


def pickle_server_decode(msg):
    """Decodes msg and returns its parts"""
    unpickler = Unpickler(BytesIO(msg))
    unpickler.find_global = server_find_global
    try:
        # PyPy, zodbpickle, the non-c-accelerated version
        unpickler.find_class = server_find_global
    except AttributeError:
        pass

    try:
        return unpickler.load()  # msgid, flags, name, args
    except:  # NOQA: E722 bare except
        logger.error("can't decode message: %s" % short_repr(msg))
        raise


def server_default(obj):
    if isinstance(obj, Exception):
        return reduce_exception(obj)
    else:
        return obj


def reduce_exception(exc):
    class_ = exc.__class__
    class_ = f'{class_.__module__}.{class_.__name__}'
    return class_, exc.__dict__ or exc.args


_globals = globals()
_silly = ('__doc__',)

exception_type_type = type(Exception)

_SAFE_MODULE_NAMES = (
    'ZopeUndo.Prefix', 'zodbpickle',
    'builtins', 'copy_reg', '__builtin__',
)


def find_global(module, name):
    """Helper for message unpickler"""
    try:
        m = __import__(module, _globals, _globals, _silly)
    except ImportError as msg:
        raise ImportError(f'import error {module}: {msg}')

    try:
        r = getattr(m, name)
    except AttributeError:
        raise ImportError(f'module {module} has no global {name}')

    safe = getattr(r, '__no_side_effects__', 0)
    if safe:
        return r

    # TODO:  is there a better way to do this?
    if type(r) == exception_type_type and issubclass(r, Exception):
        return r

    raise ImportError(f'Unsafe global: {module}.{name}')


def server_find_global(module, name):
    """Helper for message unpickler"""
    if module not in _SAFE_MODULE_NAMES:
        raise ImportError(f'Module not allowed: {module}')

    try:
        m = __import__(module, _globals, _globals, _silly)
    except ImportError as msg:
        raise ImportError(f'import error {module}: {msg}')

    try:
        r = getattr(m, name)
    except AttributeError:
        raise ImportError(f'module {module} has no global {name}')

    return r



# cython: language_level=3
"""``cython`` implementation of ``futures.py``.

Please see its docstring for details.
"""

import asyncio


cdef object CancelledError = asyncio.CancelledError
cdef object InvalidStateError = asyncio.InvalidStateError
cdef object get_event_loop = asyncio.get_event_loop
from threading import Event
from time import sleep


cdef enum State:
    PENDING = 0
    RESULT = 1
    EXCEPTION = 2
    CANCELLED = 3


cdef class Future:
    """Minimal mostly ``asyncio`` compatible future.

    In contrast to an ``asyncio`` future,
    callbacks are called immediately, not scheduled;
    their context is ignored.
    """
    cdef public object _asyncio_future_blocking

    cdef public object _loop
    cdef State state
    cdef object _result
    cdef list callbacks
    cdef bint _result_retrieved

    def __init__(self, loop=None):
        self._asyncio_future_blocking = False
        self._loop = loop if loop is not None else get_event_loop()
        self.state = PENDING
        self._result = None
        self.callbacks = []
        self._result_retrieved = False

    def get_loop(self):
        return self._loop

    cpdef cancel(self, msg=None):
        """cancel the future if not done.

        Return ``True``, if really cancelled.
        """
        if self.state:
            return False
        self.state = CANCELLED
        self._result = CancelledError()  if msg is None  else \
                       CancelledError(msg)
        self.call_callbacks()
        return True

    def cancelled(self):
        return self.state == CANCELLED

    def done(self):
        return self.state

    cpdef result(self):
        if self.state == PENDING:
            raise InvalidStateError("not done")
        self._result_retrieved = True
        if self.state == RESULT:
            return self._result
        else:
            raise self._result

    cpdef exception(self):
        if self.state == PENDING:
            raise InvalidStateError("not done")
        self._result_retrieved = True
        if self.state == RESULT:
            return None
        else:
            return self._result

    cpdef add_done_callback(self, cb, context=None):
        if not self.state or self.callbacks:
            self.callbacks.append(cb)
        else:
            cb(self)

    cpdef remove_done_callback(self, cb):
        if self.state and self.callbacks:
            raise NotImplementedError("cannot remove callbacks when done")
        flt = [c for c in self.callbacks if c != cb]
        cdef int rv = len(self.callbacks) - len(flt)
        if rv:
            self.callbacks[:] = flt
        return rv

    cdef call_callbacks(self):
        for cb in self.callbacks:  # allows ``callbacks`` to grow
            try:
                cb(self)
            except (SystemExit, KeyboardInterrupt):
                raise
            except BaseException as exc:
                self._loop.call_exception_handler({
                        'message': 'Exception in callback %s' % (cb,),
                        'exception': exc,
                    })

        del self.callbacks[:]

    cpdef set_result(self, result):
        if self.state:
            raise InvalidStateError("already done")
        self.state = RESULT
        self._result = result
        self.call_callbacks()

    cpdef set_exception(self, exc):
        if self.state:
            raise InvalidStateError("already done")
        if isinstance(exc, type):
            exc = exc()
        self.state = EXCEPTION
        self._result = exc
        self.call_callbacks()

    # py3 accesses ._exception directly
    @property
    def _exception(self):
        if self.state != EXCEPTION:
            return None
        return self._result

    def __await__(self):
        if not self.state:
            self._asyncio_future_blocking = True
            yield self
        return self.result()

    __iter__ = __await__

    def __str__(self):
        cls = self.__class__
        info = [cls.__module__ + "." + cls.__name__,
                ("PENDING", "RESULT", "EXCEPTION", "CANCELLED")[self.state],
                self._result,
                self.callbacks]
        return " ".join(str(x) for x in info)

    def __del__(self):
        if self.state == EXCEPTION  and  not self._result_retrieved:
            self._loop.call_exception_handler({
                'message': "%s exception was never retrieved" % self.__class__.__name__,
                'exception': self._result,
                'future': self,
            })


cdef class ConcurrentFuture(Future):
    """A future threads can wait on.

    Note: this differs from concurrent.future.Future - hereby ConcurrentFuture
    is generally _not_ concurrent - only .result() is allowed to be called from
    different threads and provides semantic similar to concurrent.future.Future.
    """
    cdef object completed

    def __init__(self, loop=False):
        Future.__init__(self, loop=loop)
        self.completed = Event()
        self.add_done_callback(self._complete)

    cpdef _complete(self, unused):
        self.completed.set()
        switch_thread()

    cpdef result(self, timeout=None):
        """result waits till the future is done and returns its result.

        If the future isn't done in specified time TimeoutError(*) is raised.

        (*) NOTE: it is asyncio.TimeoutError, not concurrent.futures.TimeoutError,
            which is raised for uniformity.
        """
        if not self.completed.wait(timeout):
            raise asyncio.TimeoutError()
        return Future.result(self)


IF UNAME_SYSNAME == "Windows":
    cpdef switch_thread():
        sleep(1e-6)
ELSE:
    cdef extern from "<sys/select.h>" nogil:
        ctypedef struct fd_set
        cdef struct timeval:
           long tv_sec
           unsigned long tv_usec
        int select(int no, fd_set *rd, fd_set *wr, fd_set *ex,
                   timeval *to) nogil
    cpdef switch_thread():
        cdef timeval timeout
        timeout.tv_sec = 0; timeout.tv_usec = 1
        with nogil:
            select(0, NULL, NULL, NULL, &timeout)


cdef class CoroutineExecutor:
    """Execute a coroutine on behalf of a task.

    No context support.
    """
    cdef object coro  # executed coroutine
    cdef object task   # associated task
    cdef object awaiting  # future we are waiting for
    cdef bint   cancel_requested  # request to be canceled from .cancel
    cdef object cancel_msg  # ^^^ cancellation message

    def __init__(self, task, coro):
        """execute *coro* on behalf of *task*."""
        self.task = task  # likely introduces a reference cycle
        self.coro = coro
        self.awaiting = None
        self.cancel_requested = False
        self.cancel_msg = None

    cpdef step(self):
        await_result = None  # with what to wakeup suspended await
        await_resexc = False # is it exception?
        awaiting = self.awaiting
        if awaiting is not None:
            self.awaiting = None
            assert awaiting.done()
            try:
                await_result = awaiting.result()
            except BaseException as e:
                await_result = e
                await_resexc = True
        if self.cancel_requested:
            await_result = CancelledError()  if self.cancel_msg is None  else \
                           CancelledError(self.cancel_msg)
            await_resexc = True
            self.cancel_requested = False
            self.cancel_msg = None
        try:
            if not await_resexc:
                result = self.coro.send(await_result)
            else:
                result = self.coro.throw(await_result)
        except BaseException as e:
            # we are done
            task = self.task
            self.task = None  # break reference cycle
            if isinstance(e, StopIteration):
                task.set_result(e.value)
            elif isinstance(e, CancelledError):
                if len(e.args) == 0:
                    msg = getattr(awaiting, '_cancel_message', None)  # see _cancel_future
                elif len(e.args) == 1:
                    msg = e.args[0]
                else:
                    msg = e.args
                task._cancel(msg)
            else:
                task.set_exception(e)
                if isinstance(e, (KeyboardInterrupt, SystemExit)):
                    raise
        else:
            await_next = None
            # yielded Future - wait on it
            blocking = getattr(result, '_asyncio_future_blocking', None)
            if blocking is not None:
                result._asyncio_future_blocking = False
                await_next = result

            # bad await
            else:
                await_next = Future(self.task.get_loop())
                await_next.set_exception(
                        RuntimeError("Task got bad await: %r" % (result,)))

            if self.cancel_requested:
                _cancel_future(await_next, self.cancel_msg)
                self.cancel_requested = False
                self.cancel_msg = None

            self.awaiting = await_next
            await_next.add_done_callback(self.wakeup)

            awaiting = None
            await_next = None

    cpdef wakeup(self, unused):
        self.step()

    cpdef cancel(self, msg):
        """cancel requests cancellation of the coroutine.

        It is safe to call cancel only from the same thread where coroutine is executed.
        """
        awaiting = self.awaiting
        if awaiting is not None:
            if _cancel_future(awaiting, msg):
                return True
        self.cancel_msg = msg
        self.cancel_requested = True
        task = self.task
        if task is None or task.done():
            return False
        return True

# _cancel_future cancels future fut with message msg.
# if fut does not support cancelling with message, the message is saved in fut._cancel_message .
cdef _cancel_future(fut, msg):
    try:
        return fut.cancel(msg)
    except TypeError:
        # on py3 < 3.9 Future.cancel does not accept msg
        _ = fut.cancel()
        fut._cancel_message = msg
        return _


cdef class AsyncTask(Future):
    """Simplified ``asyncio.Task``.

    Steps are not scheduled but executed immediately; the context is ignored.
    """
    cdef CoroutineExecutor executor

    def __init__(self, coro, loop=None):
        Future.__init__(self, loop=loop)
        self.executor = CoroutineExecutor(self, coro)  # reference cycle
        self.executor.step()

    cpdef cancel(self, msg=None):
        """external cancel request."""
        return self.executor.cancel(msg)

    def _cancel(self, msg):
        """internal cancel request."""
        return Future.cancel(self, msg)


cdef class ConcurrentTask(ConcurrentFuture):
    """Task reporting to ``ConcurrentFuture``.

    Steps are not scheduled but executed immediately; the context is ignored.
    Cancel can be used only from IO thread.
    """
    cdef CoroutineExecutor executor

    def __init__(self, coro, loop):
        ConcurrentFuture.__init__(self, loop=loop)
        self.executor = CoroutineExecutor(self, coro)  # reference cycle
        loop.call_soon_threadsafe(self._start)

    cpdef _start(self):
        self.executor.step()

    cpdef cancel(self, msg=None):
        """external cancel request.

        cancel requests cancellation of the task.
        it is safe to call cancel only from IO thread.
        """
        return self.executor.cancel(msg)

    def _cancel(self, msg):
        """internal cancel request."""
        return ConcurrentFuture.cancel(self, msg)


# @coroutine - only py implementtion


run_coroutine_threadsafe = ConcurrentTask



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""A TransactionBuffer store transaction updates until commit or abort.

A transaction may generate enough data that it is not practical to
always hold pending updates in memory.  Instead, a TransactionBuffer
is used to store the data until a commit or abort.
"""

# A faster implementation might store trans data in memory until it
# reaches a certain size.

import tempfile

from zodbpickle.pickle import Pickler
from zodbpickle.pickle import Unpickler


class TransactionBuffer:

    # The TransactionBuffer is used by client storage to hold update
    # data until the tpc_finish().  It is only used by a single
    # thread, because only one thread can be in the two-phase commit
    # at one time.

    def __init__(self, connection_generation):
        self.connection_generation = connection_generation
        self.file = tempfile.TemporaryFile(suffix=".tbuf")
        self.count = 0
        self.size = 0
        self.blobs = []
        # It's safe to use a fast pickler because the only objects
        # stored are builtin types -- strings or None.
        self.pickler = Pickler(self.file, 1)
        self.pickler.fast = 1
        self.server_resolved = set()  # {oid}
        self.client_resolved = {}  # {oid -> buffer_record_number}
        self.exception = None

    def close(self):
        self.file.close()

    def store(self, oid, data):
        """Store oid, version, data for later retrieval"""
        self.pickler.dump((oid, data))
        self.count += 1
        # Estimate per-record cache size
        self.size = self.size + (data and len(data) or 0) + 31

    def resolve(self, oid, data):
        """Record client-resolved data
        """
        self.store(oid, data)
        self.client_resolved[oid] = self.count - 1

    def server_resolve(self, oid):
        self.server_resolved.add(oid)

    def storeBlob(self, oid, blobfilename):
        self.blobs.append((oid, blobfilename))

    def __iter__(self):
        self.file.seek(0)
        unpickler = Unpickler(self.file)
        server_resolved = self.server_resolved
        client_resolved = self.client_resolved

        # Gaaaa, this is awkward. There can be entries in serials that
        # aren't in the buffer, because undo.  Entries can be repeated
        # in the buffer, because ZODB. (Maybe this is a bug now, but
        # it may be a feature later.

        seen = set()
        for i in range(self.count):
            oid, data = unpickler.load()
            if client_resolved.get(oid, i) == i:
                seen.add(oid)
                yield oid, data, oid in server_resolved

        # We may have leftover oids because undo
        for oid in server_resolved:
            if oid not in seen:
                yield oid, None, True



##############################################################################
#
# Copyright (c) 2001, 2002, 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Start the ZEO storage server.

Usage: %s [-C URL] [-a ADDRESS] [-f FILENAME] [-h]

Options:
-C/--configure URL -- configuration file or URL
-a/--address ADDRESS -- server address of the form PORT, HOST:PORT, or PATH
                        (a PATH must contain at least one "/")
-f/--filename FILENAME -- filename for FileStorage
-t/--timeout TIMEOUT -- transaction timeout in seconds (default no timeout)
-h/--help -- print this usage message and exit
--pid-file PATH -- relative path to output file containing this process's pid;
                   default $(INSTANCE_HOME)/var/ZEO.pid but only if envar
                   INSTANCE_HOME is defined

Unless -C is specified, -a and -f are required.
"""

# The code here is designed to be reused by other, similar servers.

import logging
import os
import signal
import socket
import sys

import ZConfig.datatypes
from zdaemon.zdoptions import ZDOptions


logger = logging.getLogger('ZEO.runzeo')
_pid = str(os.getpid())


def log(msg, level=logging.INFO, exc_info=False):
    """Internal: generic logging function."""
    message = f'({_pid}) {msg}'
    logger.log(level, message, exc_info=exc_info)


def parse_binding_address(arg):
    # Caution:  Not part of the official ZConfig API.
    obj = ZConfig.datatypes.SocketBindingAddress(arg)
    return obj.family, obj.address


def windows_shutdown_handler():
    # Called by the signal mechanism on Windows to perform shutdown.
    import asyncore
    asyncore.close_all()


class ZEOOptionsMixin:

    storages = None

    def handle_address(self, arg):
        self.family, self.address = parse_binding_address(arg)

    def handle_filename(self, arg):
        from ZODB.config import FileStorage  # That's a FileStorage *opener*!

        class FSConfig:
            def __init__(self, name, path):
                self._name = name
                self.path = path
                self.stop = None

            def getSectionName(self):
                return self._name

        if not self.storages:
            self.storages = []
        name = str(1 + len(self.storages))
        conf = FileStorage(FSConfig(name, arg))
        self.storages.append(conf)

    testing_exit_immediately = False

    def handle_test(self, *args):
        self.testing_exit_immediately = True

    def add_zeo_options(self):
        self.add(None, None, None, "test", self.handle_test)
        self.add(None, None, "a:", "address=", self.handle_address)
        self.add(None, None, "f:", "filename=", self.handle_filename)
        self.add("family", "zeo.address.family")
        self.add("address", "zeo.address.address",
                 required="no server address specified; use -a or -C")
        self.add("read_only", "zeo.read_only", default=0)
        self.add("client_conflict_resolution",
                 "zeo.client_conflict_resolution",
                 default=0)
        self.add("msgpack", "zeo.msgpack", default=0)
        self.add("invalidation_queue_size", "zeo.invalidation_queue_size",
                 default=100)
        self.add("invalidation_age", "zeo.invalidation_age")
        self.add("transaction_timeout", "zeo.transaction_timeout",
                 "t:", "timeout=", float)
        self.add('pid_file', 'zeo.pid_filename',
                 None, 'pid-file=')
        self.add("ssl", "zeo.ssl")


class ZEOOptions(ZDOptions, ZEOOptionsMixin):

    __doc__ = __doc__

    logsectionname = "eventlog"
    schemadir = os.path.dirname(__file__)

    def __init__(self):
        ZDOptions.__init__(self)
        self.add_zeo_options()
        self.add("storages", "storages",
                 required="no storages specified; use -f or -C")

    def realize(self, *a, **k):
        ZDOptions.realize(self, *a, **k)
        nunnamed = [s for s in self.storages if s.name is None]
        if nunnamed:
            if len(nunnamed) > 1:
                return self.usage("No more than one storage may be unnamed.")
            if [s for s in self.storages if s.name == '1']:
                return self.usage(
                    "Can't have an unnamed storage and a storage named 1.")
            for s in self.storages:
                if s.name is None:
                    s.name = '1'
                    break


class ZEOServer:

    def __init__(self, options):
        self.options = options
        self.server = None

    def main(self):
        self.setup_default_logging()
        self.check_socket()
        self.clear_socket()
        self.make_pidfile()
        try:
            self.open_storages()
            self.setup_signals()
            self.create_server()
            self.loop_forever()
        finally:
            self.close_server()
            self.clear_socket()
            self.remove_pidfile()

    def setup_default_logging(self):
        if self.options.config_logger is not None:
            return
        # No log file is configured; default to stderr.
        root = logging.getLogger()
        root.setLevel(logging.INFO)
        fmt = logging.Formatter(
                "------\n%(asctime)s %(levelname)s %(name)s %(message)s",
                "%Y-%m-%dT%H:%M:%S")
        handler = logging.StreamHandler()
        handler.setFormatter(fmt)
        root.addHandler(handler)

    def check_socket(self):
        if isinstance(self.options.address, tuple) and \
           self.options.address[1] is None:
            self.options.address = self.options.address[0], 0
            return

        if self.can_connect(self.options.family, self.options.address):
            self.options.usage("address %s already in use" %
                               repr(self.options.address))

    def can_connect(self, family, address):
        s = socket.socket(family, socket.SOCK_STREAM)
        try:
            s.connect(address)
        except OSError:
            return 0
        else:
            s.close()
            return 1

    def clear_socket(self):
        if isinstance(self.options.address, str):
            try:
                os.unlink(self.options.address)
            except OSError:
                pass

    def open_storages(self):
        self.storages = {}
        for opener in self.options.storages:
            log("opening storage %r using %s"
                % (opener.name, opener.__class__.__name__))
            self.storages[opener.name] = opener.open()

    def setup_signals(self):
        """Set up signal handlers.

        The signal handler for SIGFOO is a method handle_sigfoo().
        If no handler method is defined for a signal, the signal
        action is not changed from its initial value.  The handler
        method is called without additional arguments.
        """
        if os.name != "posix":
            if os.name == "nt":
                self.setup_win32_signals()
            return
        if hasattr(signal, 'SIGXFSZ'):
            signal.signal(signal.SIGXFSZ, signal.SIG_IGN)  # Special case
        init_signames()
        for sig, name in signames.items():
            method = getattr(self, "handle_" + name.lower(), None)
            if method is not None:
                def wrapper(sig_dummy, frame_dummy, method=method):
                    method()
                signal.signal(sig, wrapper)

    def setup_win32_signals(self):
        # Borrow the Zope Signals package win32 support, if available.
        # Signals does a check/log for the availability of pywin32.
        try:
            import Signals.Signals
        except ImportError:
            logger.debug("Signals package not found. "
                         "Windows-specific signal handler "
                         "will *not* be installed.")
            return
        SignalHandler = Signals.Signals.SignalHandler
        if SignalHandler is not None:  # may be None if no pywin32.
            SignalHandler.registerHandler(signal.SIGTERM,
                                          windows_shutdown_handler)
            SignalHandler.registerHandler(signal.SIGINT,
                                          windows_shutdown_handler)
            SIGUSR2 = 12  # not in signal module on Windows.
            SignalHandler.registerHandler(SIGUSR2, self.handle_sigusr2)

    def create_server(self):
        self.server = create_server(self.storages, self.options)

    def loop_forever(self):
        if self.options.testing_exit_immediately:
            print("testing exit immediately")
            sys.stdout.flush()  # ensure truthful output order
        else:
            self.server.loop()

    def close_server(self):
        if self.server is not None:
            self.server.close()
            if getattr(self.options, "testing_exit_immediately", False):
                acceptor = self.server.acceptor
                if hasattr(acceptor, "event_loop"):
                    # usually, this happens automatically - but not for testing
                    acceptor.event_loop.close()

    def handle_sigterm(self):
        log("terminated by SIGTERM")
        sys.exit(0)

    def handle_sigint(self):
        log("terminated by SIGINT")
        sys.exit(0)

    def handle_sighup(self):
        log("restarted by SIGHUP")
        sys.exit(1)

    def handle_sigusr2(self):
        # log rotation signal - do the same as Zope 2.7/2.8...
        if self.options.config_logger is None or \
           os.name not in ("posix", "nt"):
            log("received SIGUSR2, but it was not handled!",
                level=logging.WARNING)
            return

        loggers = [self.options.config_logger]

        if os.name == "posix":
            for logger in loggers:
                logger.reopen()
            log("Log files reopened successfully", level=logging.INFO)
        else:  # nt - same rotation code as in Zope's Signals/Signals.py
            for logger in loggers:
                for factory in logger.handler_factories:
                    handler = factory()
                    if hasattr(handler, 'rotate') and callable(handler.rotate):
                        handler.rotate()
            log("Log files rotation complete", level=logging.INFO)

    def _get_pidfile(self):
        pidfile = self.options.pid_file
        # 'pidfile' is marked as not required.
        if not pidfile:
            # Try to find a reasonable location if the pidfile is not
            # set. If we are running in a Zope environment, we can
            # safely assume INSTANCE_HOME.
            instance_home = os.environ.get("INSTANCE_HOME")
            if not instance_home:
                # If all our attempts failed, just log a message and
                # proceed.
                logger.debug("'pidfile' option not set, and 'INSTANCE_HOME' "
                             "environment variable could not be found. "
                             "Cannot guess pidfile location.")
                return
            self.options.pid_file = os.path.join(instance_home,
                                                 "var", "ZEO.pid")

    def make_pidfile(self):
        if not self.options.read_only:
            self._get_pidfile()
            pidfile = self.options.pid_file
            if pidfile is None:
                return
            pid = os.getpid()
            try:
                if os.path.exists(pidfile):
                    os.unlink(pidfile)
                f = open(pidfile, 'w')
                print(pid, file=f)
                f.close()
                log("created PID file '%s'" % pidfile)
            except OSError:
                logger.error("PID file '%s' cannot be opened" % pidfile)

    def remove_pidfile(self):
        if not self.options.read_only:
            pidfile = self.options.pid_file
            if pidfile is None:
                return
            try:
                if os.path.exists(pidfile):
                    os.unlink(pidfile)
                    log("removed PID file '%s'" % pidfile)
            except OSError:
                logger.error("PID file '%s' could not be removed" % pidfile)


def create_server(storages, options):
    from .StorageServer import StorageServer
    return StorageServer(
        options.address,
        storages,
        read_only=options.read_only,
        client_conflict_resolution=options.client_conflict_resolution,
        msgpack=(options.msgpack if isinstance(options.msgpack, bool)
                 else os.environ.get('ZEO_MSGPACK')),
        invalidation_queue_size=options.invalidation_queue_size,
        invalidation_age=options.invalidation_age,
        transaction_timeout=options.transaction_timeout,
        ssl=options.ssl)


# Signal names

signames = None


def signame(sig):
    """Return a symbolic name for a signal.

    Return "signal NNN" if there is no corresponding SIG name in the
    signal module.
    """

    if signames is None:
        init_signames()
    return signames.get(sig) or "signal %d" % sig


def init_signames():
    global signames
    signames = {}
    for name, sig in signal.__dict__.items():
        k_startswith = getattr(name, "startswith", None)
        if k_startswith is None:
            continue
        if k_startswith("SIG") and not k_startswith("SIG_"):
            signames[sig] = name


# Main program

def main(args=None):
    options = ZEOOptions()
    options.realize(args)
    s = ZEOServer(options)
    s.main()


def run(args):
    options = ZEOOptions()
    options.realize(args)
    s = ZEOServer(options)
    s.run()


if __name__ == "__main__":
    main()



"""Testing helpers
"""
import ZEO.StorageServer

from ..asyncio.server import best_protocol_version


class ServerProtocol:

    method = ('register', )

    def __init__(self, zs,
                 protocol_version=best_protocol_version,
                 addr='test-address'):
        self.calls = []
        self.addr = addr
        self.zs = zs
        self.protocol_version = protocol_version
        zs.notify_connected(self)

    closed = False

    def close(self):
        if not self.closed:
            self.closed = True
            self.zs.notify_disconnected()

    def call_soon_threadsafe(self, func, *args):
        func(*args)

    def async_(self, *args):
        self.calls.append(args)

    async_threadsafe = async_


class StorageServer:
    """Create a client interface to a StorageServer.

    This is for testing StorageServer. It interacts with the storgr
    server through its network interface, but without creating a
    network connection.
    """

    def __init__(self, test, storage,
                 protocol_version=b'Z' + best_protocol_version,
                 **kw):
        self.test = test
        self.storage_server = ZEO.StorageServer.StorageServer(
            None, {'1': storage}, **kw)
        self.zs = self.storage_server.create_client_handler()
        self.protocol = ServerProtocol(self.zs,
                                       protocol_version=protocol_version)
        self.zs.register('1', kw.get('read_only', False))

    def assert_calls(self, test, *argss):
        if argss:
            for args in argss:
                test.assertEqual(self.protocol.calls.pop(0), args)
        else:
            test.assertEqual(self.protocol.calls, ())

    def unpack_result(self, result):
        """For methods that return Result objects, unwrap the results
        """
        result, callback = result.args
        callback()
        return result



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Monitor behavior of ZEO server and record statistics.
"""

import time


zeo_version = 'unknown'
try:
    import pkg_resources
except ImportError:
    pass
else:
    zeo_dist = pkg_resources.working_set.find(
        pkg_resources.Requirement.parse('ZODB3')
        )
    if zeo_dist is not None:
        zeo_version = zeo_dist.version


class StorageStats:
    """Per-storage usage statistics."""

    def __init__(self, connections=None):
        self.connections = connections
        self.loads = 0
        self.stores = 0
        self.commits = 0
        self.aborts = 0
        self.active_txns = 0
        self.lock_time = None
        self.conflicts = 0
        self.conflicts_resolved = 0
        self.start = time.ctime()

    @property
    def clients(self):
        return len(self.connections)

    def parse(self, s):
        # parse the dump format
        lines = s.split("\n")
        for line in lines:
            field, value = line.split(":", 1)
            if field == "Server started":
                self.start = value
            elif field == "Clients":
                # Hack because we use this both on the server and on
                # the client where there are no connections.
                self.connections = [0] * int(value)
            elif field == "Clients verifying":
                self.verifying_clients = int(value)
            elif field == "Active transactions":
                self.active_txns = int(value)
            elif field == "Commit lock held for":
                # This assumes
                self.lock_time = time.time() - int(value)
            elif field == "Commits":
                self.commits = int(value)
            elif field == "Aborts":
                self.aborts = int(value)
            elif field == "Loads":
                self.loads = int(value)
            elif field == "Stores":
                self.stores = int(value)
            elif field == "Conflicts":
                self.conflicts = int(value)
            elif field == "Conflicts resolved":
                self.conflicts_resolved = int(value)

    def dump(self, f):
        print("Server started:", self.start, file=f)
        print("Clients:", self.clients, file=f)
        print("Clients verifying:", self.verifying_clients, file=f)
        print("Active transactions:", self.active_txns, file=f)
        if self.lock_time:
            howlong = time.time() - self.lock_time
            print("Commit lock held for:", int(howlong), file=f)
        print("Commits:", self.commits, file=f)
        print("Aborts:", self.aborts, file=f)
        print("Loads:", self.loads, file=f)
        print("Stores:", self.stores, file=f)
        print("Conflicts:", self.conflicts, file=f)
        print("Conflicts resolved:", self.conflicts_resolved, file=f)



##############################################################################
#
# Copyright Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import doctest
import logging
import pprint
import re
import sys
import unittest

import transaction
import ZODB.blob
import ZODB.FileStorage
import ZODB.tests.util
import ZODB.utils
from zope.testing import renormalizing
from zope.testing import setupstack

import ZEO.StorageServer
import ZEO.tests.servertesting


def proper_handling_of_blob_conflicts():
    r"""

Conflict errors weren't properly handled when storing blobs, the
result being that the storage was left in a transaction.

We originally saw this when restarting a blob transaction, although
it doesn't really matter.

Set up the storage with some initial blob data.

    >>> fs = ZODB.FileStorage.FileStorage('t.fs', blob_dir='t.blobs')
    >>> db = ZODB.DB(fs)
    >>> conn = db.open()
    >>> conn.root.b = ZODB.blob.Blob(b'x')
    >>> transaction.commit()

Get the oid and first serial. We'll use the serial later to provide
out-of-date data.

    >>> oid = conn.root.b._p_oid
    >>> serial = conn.root.b._p_serial
    >>> with conn.root.b.open('w') as file:
    ...     _ = file.write(b'y')
    >>> transaction.commit()
    >>> data = fs.load(oid)[0]

Create the server:

    >>> server = ZEO.tests.servertesting.StorageServer('x', {'1': fs})

And an initial client.

    >>> zs1 = ZEO.tests.servertesting.client(server, 1)
    >>> zs1.tpc_begin('0', '', '', {})
    >>> zs1.storea(ZODB.utils.p64(99), ZODB.utils.z64, b'x', '0')
    >>> _ = zs1.vote('0') # doctest: +ELLIPSIS

In a second client, we'll try to commit using the old serial. This
will conflict. It will be blocked at the vote call.

    >>> zs2 = ZEO.tests.servertesting.client(server, 2)
    >>> zs2.tpc_begin('1', '', '', {})
    >>> zs2.storeBlobStart()
    >>> zs2.storeBlobChunk(b'z')
    >>> zs2.storeBlobEnd(oid, serial, data, '1')
    >>> delay = zs2.vote('1')

    >>> class Sender:
    ...     def send_reply(self, id, reply):
    ...         print('reply', id, reply)
    ...     def send_error(self, id, err):
    ...         print('error', id, err)
    >>> delay.set_sender(1, Sender())

    >>> logger = logging.getLogger('ZEO')
    >>> handler = logging.StreamHandler(sys.stdout)
    >>> logger.setLevel(logging.INFO)
    >>> logger.addHandler(handler)

Now, when we abort the transaction for the first client. The second
client will be restarted.  It will get a conflict error, that is
raised to the client:

    >>> zs1.tpc_abort('0') # doctest: +ELLIPSIS
    error 1 database conflict error ...

The transaction is aborted by the server:

    >>> fs.tpc_transaction() is None
    True

    >>> zs2.connected
    True

    >>> logger.setLevel(logging.NOTSET)
    >>> logger.removeHandler(handler)
    >>> zs2.tpc_abort('1')
    >>> fs.close()

    >>> server.close()
    """


def proper_handling_of_errors_in_restart():
    r"""

It's critical that if there is an error in vote that the
storage isn't left in tpc.

    >>> fs = ZODB.FileStorage.FileStorage('t.fs', blob_dir='t.blobs')
    >>> server = ZEO.tests.servertesting.StorageServer('x', {'1': fs})

And an initial client.

    >>> zs1 = ZEO.tests.servertesting.client(server, 1)
    >>> zs1.tpc_begin('0', '', '', {})
    >>> zs1.storea(ZODB.utils.p64(99), ZODB.utils.z64, b'x', '0')

Intentionally break zs1:

    >>> zs1._store = lambda : None
    >>> _ = zs1.vote('0') # doctest: +ELLIPSIS +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    TypeError: <lambda>() takes no arguments (3 given)

We're not in a transaction:

    >>> fs.tpc_transaction() is None
    True

We can start another client and get the storage lock.

    >>> zs1 = ZEO.tests.servertesting.client(server, 1)
    >>> zs1.tpc_begin('1', '', '', {})
    >>> zs1.storea(ZODB.utils.p64(99), ZODB.utils.z64, b'x', '1')
    >>> _ = zs1.vote('1') # doctest: +ELLIPSIS

    >>> zs1.tpc_finish('1').set_sender(0, zs1.connection)

    >>> fs.close()
    >>> server.close()
    """


def errors_in_vote_should_clear_lock():
    """

So, we arrange to get an error in vote:

    >>> import ZODB.MappingStorage
    >>> vote_should_fail = True
    >>> class MappingStorage(ZODB.MappingStorage.MappingStorage):
    ...     def tpc_vote(*args):
    ...         if vote_should_fail:
    ...             raise ValueError
    ...         return ZODB.MappingStorage.MappingStorage.tpc_vote(*args)

    >>> server = ZEO.tests.servertesting.StorageServer(
    ...      'x', {'1': MappingStorage()})
    >>> zs = ZEO.tests.servertesting.client(server, 1)
    >>> zs.tpc_begin('0', '', '', {})
    >>> zs.storea(ZODB.utils.p64(99), ZODB.utils.z64, 'x', '0')

    >>> zs.vote('0')
    Traceback (most recent call last):
    ...
    ValueError

When we do, the storage server's transaction lock shouldn't be held:

    >>> zs.lock_manager.locked is not None
    False

Of course, if vote suceeds, the lock will be held:

    >>> vote_should_fail = False
    >>> zs.tpc_begin('1', '', '', {})
    >>> zs.storea(ZODB.utils.p64(99), ZODB.utils.z64, 'x', '1')
    >>> _ = zs.vote('1') # doctest: +ELLIPSIS

    >>> zs.lock_manager.locked is not None
    True

    >>> zs.tpc_abort('1')

    >>> server.close()
    """


def some_basic_locking_tests():
    r"""

    >>> itid = 0
    >>> def start_trans(zs):
    ...     global itid
    ...     itid += 1
    ...     tid = str(itid)
    ...     zs.tpc_begin(tid, '', '', {})
    ...     zs.storea(ZODB.utils.p64(99), ZODB.utils.z64, 'x', tid)
    ...     return tid

    >>> server = ZEO.tests.servertesting.StorageServer()

    >>> handler = logging.StreamHandler(sys.stdout)
    >>> handler.setFormatter(logging.Formatter(
    ...     '%(name)s %(levelname)s\n%(message)s'))
    >>> logging.getLogger('ZEO').addHandler(handler)
    >>> logging.getLogger('ZEO').setLevel(logging.DEBUG)

Work around the fact that ZODB registers level names backwards:

    >>> import logging
    >>> from ZODB.loglevels import BLATHER
    >>> logging.addLevelName(BLATHER, "BLATHER")

We start a transaction and vote, this leads to getting the lock.

    >>> zs1 = ZEO.tests.servertesting.client(server, '1')
    ZEO.asyncio.base INFO
    Connected ZEO.asyncio.server.ServerProtocol('test-addr-1')
    ZEO.asyncio.server INFO
    received handshake 'Z5'
    >>> tid1 = start_trans(zs1)
    >>> resolved1 = zs1.vote(tid1) # doctest: +ELLIPSIS
    ZEO.StorageServer DEBUG
    (test-addr-1) ('1') lock: transactions waiting: 0
    ZEO.StorageServer BLATHER
    (test-addr-1) Preparing to commit transaction: 1 objects, ... bytes

If another client tried to vote, it's lock request will be queued and
a delay will be returned:

    >>> zs2 = ZEO.tests.servertesting.client(server, '2')
    ZEO.asyncio.base INFO
    Connected ZEO.asyncio.server.ServerProtocol('test-addr-2')
    ZEO.asyncio.server INFO
    received handshake 'Z5'
    >>> tid2 = start_trans(zs2)
    >>> delay = zs2.vote(tid2)
    ZEO.StorageServer DEBUG
    (test-addr-2) ('1') queue lock: transactions waiting: 1

    >>> delay.set_sender(0, zs2.connection)

When we end the first transaction, the queued vote gets the lock.

    >>> zs1.tpc_abort(tid1) # doctest: +ELLIPSIS
    ZEO.StorageServer DEBUG
    (test-addr-1) ('1') unlock: transactions waiting: 1
    ZEO.StorageServer DEBUG
    (test-addr-2) ('1') lock: transactions waiting: 0
    ZEO.StorageServer BLATHER
    (test-addr-2) Preparing to commit transaction: 1 objects, ... bytes

Let's try again with the first client. The vote will be queued:

    >>> tid1 = start_trans(zs1)
    >>> delay = zs1.vote(tid1)
    ZEO.StorageServer DEBUG
    (test-addr-1) ('1') queue lock: transactions waiting: 1

If the queued transaction is aborted, it will be dequeued:

    >>> zs1.tpc_abort(tid1) # doctest: +ELLIPSIS
    ZEO.StorageServer DEBUG
    (test-addr-1) ('1') dequeue lock: transactions waiting: 0

BTW, voting multiple times will error:

    >>> zs2.vote(tid2)
    Traceback (most recent call last):
    ...
    StorageTransactionError: Already voting (locked)

    >>> tid1 = start_trans(zs1)
    >>> delay = zs1.vote(tid1)
    ZEO.StorageServer DEBUG
    (test-addr-1) ('1') queue lock: transactions waiting: 1

    >>> delay.set_sender(0, zs1.connection)

    >>> zs1.vote(tid1)
    Traceback (most recent call last):
    ...
    StorageTransactionError: Already voting (waiting)

Note that the locking activity is logged at debug level to avoid
cluttering log files, however, as the number of waiting votes
increased, so does the logging level:

    >>> clients = []
    >>> for i in range(9):
    ...     client = ZEO.tests.servertesting.client(server, str(i+10))
    ...     tid = start_trans(client)
    ...     delay = client.vote(tid)
    ...     clients.append(client)
    ZEO.asyncio.base INFO
    Connected ZEO.asyncio.server.ServerProtocol('test-addr-10')
    ZEO.asyncio.server INFO
    received handshake 'Z5'
    ZEO.StorageServer DEBUG
    (test-addr-10) ('1') queue lock: transactions waiting: 2
    ZEO.asyncio.base INFO
    Connected ZEO.asyncio.server.ServerProtocol('test-addr-11')
    ZEO.asyncio.server INFO
    received handshake 'Z5'
    ZEO.StorageServer DEBUG
    (test-addr-11) ('1') queue lock: transactions waiting: 3
    ZEO.asyncio.base INFO
    Connected ZEO.asyncio.server.ServerProtocol('test-addr-12')
    ZEO.asyncio.server INFO
    received handshake 'Z5'
    ZEO.StorageServer WARNING
    (test-addr-12) ('1') queue lock: transactions waiting: 4
    ZEO.asyncio.base INFO
    Connected ZEO.asyncio.server.ServerProtocol('test-addr-13')
    ZEO.asyncio.server INFO
    received handshake 'Z5'
    ZEO.StorageServer WARNING
    (test-addr-13) ('1') queue lock: transactions waiting: 5
    ZEO.asyncio.base INFO
    Connected ZEO.asyncio.server.ServerProtocol('test-addr-14')
    ZEO.asyncio.server INFO
    received handshake 'Z5'
    ZEO.StorageServer WARNING
    (test-addr-14) ('1') queue lock: transactions waiting: 6
    ZEO.asyncio.base INFO
    Connected ZEO.asyncio.server.ServerProtocol('test-addr-15')
    ZEO.asyncio.server INFO
    received handshake 'Z5'
    ZEO.StorageServer WARNING
    (test-addr-15) ('1') queue lock: transactions waiting: 7
    ZEO.asyncio.base INFO
    Connected ZEO.asyncio.server.ServerProtocol('test-addr-16')
    ZEO.asyncio.server INFO
    received handshake 'Z5'
    ZEO.StorageServer WARNING
    (test-addr-16) ('1') queue lock: transactions waiting: 8
    ZEO.asyncio.base INFO
    Connected ZEO.asyncio.server.ServerProtocol('test-addr-17')
    ZEO.asyncio.server INFO
    received handshake 'Z5'
    ZEO.StorageServer WARNING
    (test-addr-17) ('1') queue lock: transactions waiting: 9
    ZEO.asyncio.base INFO
    Connected ZEO.asyncio.server.ServerProtocol('test-addr-18')
    ZEO.asyncio.server INFO
    received handshake 'Z5'
    ZEO.StorageServer CRITICAL
    (test-addr-18) ('1') queue lock: transactions waiting: 10

If a client with the transaction lock disconnects, it will abort and
release the lock and one of the waiting clients will get the lock.

    >>> zs2.notify_disconnected() # doctest: +ELLIPSIS
    ZEO.StorageServer INFO
    (test-addr-...) disconnected during locked transaction
    ZEO.StorageServer CRITICAL
    (test-addr-...) ('1') unlock: transactions waiting: 10
    ZEO.StorageServer WARNING
    (test-addr-...) ('1') lock: transactions waiting: 9
    ZEO.StorageServer BLATHER
    (test-addr-...) Preparing to commit transaction: 1 objects, ... bytes

(In practice, waiting clients won't necessarily get the lock in order.)

We can find out about the current lock state, and get other server
statistics using the server_status method:

    >>> pprint.pprint(zs1.server_status(), width=40)
    {'aborts': 3,
     'active_txns': 10,
     'commits': 0,
     'conflicts': 0,
     'conflicts_resolved': 0,
     'connections': 10,
     'last-transaction': '0000000000000000',
     'loads': 0,
     'lock_time': 1272653598.693882,
     'start': 'Fri Apr 30 14:53:18 2010',
     'stores': 13,
     'timeout-thread-is-alive': 'stub',
     'waiting': 9}

If clients disconnect while waiting, they will be dequeued:

    >>> for client in clients:
    ...     client.notify_disconnected() # doctest: +ELLIPSIS
    ZEO.StorageServer INFO
    (test-addr-10) disconnected during...locked transaction
    ZEO.StorageServer WARNING
    (test-addr-10) ('1') ... lock: transactions waiting: ...

    >>> zs1.server_status()['waiting']
    0

    >>> zs1.tpc_abort(tid1)
    ZEO.StorageServer DEBUG
    (test-addr-1) ('1') unlock: transactions waiting: 0

    >>> logging.getLogger('ZEO').setLevel(logging.NOTSET)
    >>> logging.getLogger('ZEO').removeHandler(handler)
    >>> server.close()
    """


def lock_sanity_check():
    r"""
On one occasion with 3.10.0a1 in production, we had a case where a
transaction lock wasn't released properly.  One possibility, fron
scant log information, is that the server and ZEOStorage had different
ideas about whether the ZEOStorage was locked. The timeout thread
properly closed the ZEOStorage's connection, but the ZEOStorage didn't
release it's lock, presumably because it thought it wasn't locked. I'm
not sure why this happened.  I've refactored the logic quite a bit to
try to deal with this, but the consequences of this failure are so
severe, I'm adding some sanity checking when queueing lock requests.

Helper to manage transactions:

    >>> itid = 0
    >>> def start_trans(zs):
    ...     global itid
    ...     itid += 1
    ...     tid = str(itid)
    ...     zs.tpc_begin(tid, '', '', {})
    ...     zs.storea(ZODB.utils.p64(99), ZODB.utils.z64, 'x', tid)
    ...     return tid

Set up server and logging:

    >>> server = ZEO.tests.servertesting.StorageServer()

    >>> handler = logging.StreamHandler(sys.stdout)
    >>> handler.setFormatter(logging.Formatter(
    ...     '%(name)s %(levelname)s\n%(message)s'))
    >>> logging.getLogger('ZEO').addHandler(handler)
    >>> logging.getLogger('ZEO').setLevel(logging.DEBUG)

Work around the fact that ZODB registers level names backwards:

    >>> import logging
    >>> from ZODB.loglevels import BLATHER
    >>> logging.addLevelName(BLATHER, "BLATHER")

Now, we'll start a transaction, get the lock and then mark the
ZEOStorage as closed and see if trying to get a lock cleans it up:

    >>> zs1 = ZEO.tests.servertesting.client(server, '1')
    ZEO.asyncio.base INFO
    Connected ZEO.asyncio.server.ServerProtocol('test-addr-1')
    ZEO.asyncio.server INFO
    received handshake 'Z5'
    >>> tid1 = start_trans(zs1)
    >>> resolved1 = zs1.vote(tid1) # doctest: +ELLIPSIS
    ZEO.StorageServer DEBUG
    (test-addr-1) ('1') lock: transactions waiting: 0
    ZEO.StorageServer BLATHER
    (test-addr-1) Preparing to commit transaction: 1 objects, ... bytes

    >>> zs1.connection.connection_lost(None)
    ZEO.StorageServer INFO
    (test-addr-1) disconnected during locked transaction
    ZEO.StorageServer DEBUG
    (test-addr-1) ('1') unlock: transactions waiting: 0

    >>> zs2 = ZEO.tests.servertesting.client(server, '2')
    ZEO.asyncio.base INFO
    Connected ZEO.asyncio.server.ServerProtocol('test-addr-2')
    ZEO.asyncio.server INFO
    received handshake 'Z5'
    >>> tid2 = start_trans(zs2)
    >>> resolved2 = zs2.vote(tid2) # doctest: +ELLIPSIS
    ZEO.StorageServer DEBUG
    (test-addr-2) ('1') lock: transactions waiting: 0
    ZEO.StorageServer BLATHER
    (test-addr-2) Preparing to commit transaction: 1 objects, ... bytes

    >>> zs2.tpc_abort(tid2)
    ZEO.StorageServer DEBUG
    (test-addr-2) ('1') unlock: transactions waiting: 0

    >>> logging.getLogger('ZEO').setLevel(logging.NOTSET)
    >>> logging.getLogger('ZEO').removeHandler(handler)

    >>> server.close()
    """


def test_suite():
    return unittest.TestSuite((
        doctest.DocTestSuite(
            setUp=ZODB.tests.util.setUp, tearDown=setupstack.tearDown,
            checker=renormalizing.RENormalizing([
                (re.compile(r'\d+/test-addr'), ''),
                (re.compile(r"'lock_time': \d+.\d+"), 'lock_time'),
                (re.compile(r"'start': '[^\n]+'"), 'start'),
                (re.compile('ZODB.POSException.StorageTransactionError'),
                 'StorageTransactionError'),
                ]),
            ),
        ))
